import vmi
from PySide2.QtGui import Qt, QImage, QColor, QPainter


def read_png() -> QImage:
    """读取*.png文件"""
    file_name = vmi.askOpenFile('*.png')
    if file_name is not None:
        image = QImage(file_name)
        if image:
            return image


def LeftButtonPressRelease(**kwargs):
    if kwargs['picked'] is image_box:
        # 重新读取png图像
        png = read_png()
        if png is not None:
            image_box.draw(png)
    elif kwargs['picked'] is click_box:
        # 切换选项
        options = ['单击', '再击', '三击']
        options = dict(zip(options, [*options[1:], options[0]]))
        click_box.draw_text(options[click_box.text()])
    elif kwargs['picked'] is value_box:
        # 输入数值
        global value
        v = vmi.askInt(0, value, 100)
        if v is not None:
            value = vmi.askInt(0, value, 100)
            draw_value()


def NoButtonWheel(**kwargs):
    if kwargs['picked'] is value_box:
        global value
        value = min(max(value + kwargs['delta'], 0), 100)  # 控制数值范围
        draw_value()  # 更新绘图框


def draw_value():
    """自定义绘图"""
    value_box.draw_text(text='滚轮或单击：{}'.format(value))

    w, h = value_draw_box.size()[0] * view.width(), value_draw_box.size()[1] * view.height()
    painter = QPainter()
    image = QImage(w, h, value_draw_box.image().format())
    painter.begin(image)

    painter.fillRect(0, 0, w, h, QColor('white'))
    painter.fillRect(0, 0, value / 100 * w, h, QColor('lightskyblue'))

    painter.end()
    value_draw_box.draw(image)


if __name__ == '__main__':
    png = read_png()  # 读取PNG图像
    if png is None:  # 判断用户是否有效读取了模型文件
        vmi.appexit()  # 清理并退出程序

    view = vmi.View()  # 视图

    # 画一个三维物体作参照
    mesh_prop = vmi.PolyActor(view, color=[1, 1, 0.6])
    mesh_prop.setData(vmi.ccCylinder(5, 20, [0, 0, 0], [0, -1, -1]))

    # 绘图框，显示外部png图像
    # siz：视图中的显示尺寸，视图宽高的比例
    # pos：视图中的对齐位置，视图宽高的比例
    # anchor：图像中的对齐位置，图像宽高的比例
    image_box = vmi.ImageBox(view, image=png, size=[0.2, 0.4], pos=[1, 0.1], anchor=[1, 0], pickable=True)
    image_box.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease

    # 文本框，fore_color文本颜色，back_color背景颜色
    # bold：粗体
    # italic：斜体
    # underline：下划线
    # 颜色名称：https://www.w3.org/wiki/CSS/Properties/color/keywords
    text_box = vmi.TextBox(view, text='静态文本' + ' ' * 10,
                           fore_color=QColor('white'), back_color=QColor('crimson'),
                           bold=True, italic=True, underline=True,
                           size=[0.2, 0.04], pos=[0, 0.1], anchor=[0, 0])

    # 单击框，text_align组合文本对齐方式
    # 水平：Qt.AlignLeft/Qt.AlignRight/Qt.AlignHCenter
    # 竖直：Qt.AlignTop/Qt.AlignBottom/Qt.AlignVCenter/Qt.AlignBaseline
    # 水平竖直：Qt.AlignCenter
    click_box = vmi.TextBox(view, text='单击', text_align=Qt.AlignCenter,
                            size=[0.2, 0.04], pos=[0, 0.14], anchor=[0, 0], pickable=True)
    click_box.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease

    # 数值框
    value = 50
    value_box = vmi.TextBox(view, size=[0.2, 0.04], pos=[0, 0.45], anchor=[0, 0], pickable=True)
    value_box.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease
    value_box.mouse['NoButton']['Wheel'] = NoButtonWheel

    # 绘图框，自定义绘图
    value_draw_box = vmi.ImageBox(view, size=[0.2, 0.01], pos=[0, 0.49], anchor=[0, 0])

    # 初始化
    draw_value()

    view.setCamera_FitAll()  # 自动调整视图的视野范围
    vmi.appexec(view)  # 执行主窗口程序
    vmi.appexit()  # 清理并退出程序



import vmi


def read_dicom():
    dcmdir = vmi.askDirectory()  # 用户选择文件夹

    if dcmdir is not None:  # 判断用户选中了有效文件夹并点击了确认
        series_list = vmi.sortSeries(dcmdir)  # 将文件夹及其子目录包含的所有DICOM文件分类到各个系列

        if len(series_list) > 0:  # 判断该文件夹内包含有效的DICOM系列
            series = vmi.askSeries(series_list)  # 用户选择DICOM系列

            if series is not None:  # 判断用户选中了有效系列并点击了确认
                return series.read()  # 读取DICOM系列为图像数据


image = read_dicom()  # 读取DICOM图像

if image is None:  # 图像无效则退出主程序
    vmi.appexit()

view = vmi.View()  # 创建视图

slice = vmi.ImageSlice(view)  # 创建图像断层
slice.setData(image)  # 载入图像
slice.setSlicePlaneOrigin_Center()  # 设置图像断层位置居中
slice.setSlicePlaneNormal_Axial()  # 设置图像断层显示横断面
view.setCamera_FitAll()  # 重置相机适应场景大小

vmi.appexec(view)  # 执行主窗口程序
vmi.appexit()  # 清理并退出程序



import vmi
import numpy as np


def read_dicom():
    dcmdir = vmi.askDirectory()  # 用户选择文件夹

    if dcmdir is not None:  # 判断用户选中了有效文件夹并点击了确认
        series_list = vmi.sortSeries(dcmdir)  # 将文件夹及其子目录包含的所有DICOM文件分类到各个系列

        if len(series_list) > 0:  # 判断该文件夹内包含有效的DICOM系列
            series = vmi.askSeries(series_list)  # 用户选择DICOM系列

            if series is not None:  # 判断用户选中了有效系列并点击了确认
                return series.read()  # 读取DICOM系列为图像数据


def LeftButtonPressRelease(**kwargs):
    if kwargs['picked'] is button:  # 用户点击文字框
        left_top = views[0].pickPt_FocalPlane([0, 0])  # 拾取视图左上角的焦平面点
        right_top = views[0].pickPt_FocalPlane([views[0].width(), 0])  # 拾取视图右上角的焦平面点
        left_bottom = views[0].pickPt_FocalPlane([0, views[0].height()])  # 拾取视图左下角的焦平面点

        cs = views[0].cameraCS()  # 获得相机坐标系

        xl = np.linalg.norm(left_top - right_top)  # 视图左右方向的真实尺度
        yl = np.linalg.norm(left_top - left_bottom)  # 视图上下方向的真实尺度
        zl = vmi.imSize_Vt(image, cs.axis(2))  # 图像沿视图入射方向的最大包围盒尺寸

        pt = vmi.imCenter(image) - 0.5 * zl * cs.axis(2)  # 计算图像包围盒中心到近平面投影点
        cs.setOrigin(vmi.ptOnPlane(cs.origin(), pt, cs.axis(2)))  # 将坐标系原点投影到近平面

        # 重采样混叠，背景体素值-1024，平均密度投影，忽略-200以下的体素值
        image_blend = vmi.imReslice_Blend(image, cs, [xl, yl, zl], -1024, 'mean', -200)

        blend_prop.setData(image_blend)  # 载入混叠图像
        blend_prop.setColorWindow_Soft()  # 设置软组织窗
        views[1].setCamera_FitAll()  # 自动调整视图的视野范围


image = read_dicom()  # 读取DICOM路径
if image is None:
    vmi.appexit()

iso_value = vmi.askInt(-1000, 200, 3000, None, 'HU')  # 用户输入-1000HU至3000HU范围内的一个CT值
if iso_value is None:
    vmi.appexit()

views = [vmi.View(), vmi.View()]  # 左右视图

# 创建按钮，连接事件处理函数
button = vmi.TextBox(views[1], text='混叠', pickable=True, size=[0.2, 0.04], pos=[0, 0.1], anchor=[0, 0])
button.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

mesh_prop = vmi.PolyActor(views[0], color=[1, 1, 0.6])  # 左视图显示模型
mesh_prop.setData(vmi.imIsosurface(image, iso_value))  # 三维重建并显示
views[0].setCamera_Coronal()  # 设置冠状位视图
views[0].setCamera_FitAll()  # 自动调整视图的视野范围

blend_prop = vmi.ImageSlice(views[1])  # 右视图显示混叠图像

# 导入Qt用于窗口布局
from PySide2.QtWidgets import QWidget, QHBoxLayout

# 创建窗口，布局视图
widget = QWidget()
layout = QHBoxLayout(widget)
layout.addWidget(views[0])
layout.addWidget(views[1])

vmi.appexec(widget)  # 执行主程序
vmi.appexit()  # 退出主程序



import tensorflow as tf
import tensorflow.contrib as tfcontrib
from tensorflow.python.keras import layers
from tensorflow.python.keras import losses
from tensorflow.python.keras import models
from tensorflow.python.keras import backend as K
from tensorflow.python.keras import optimizers
from tensorflow.python.keras import utils
import numpy as np


def conv_block(input_tensor, num_filters):
    encoder = layers.Conv3D(num_filters, (3, 3, 3), padding='same')(input_tensor)
    encoder = layers.BatchNormalization()(encoder)
    encoder = layers.Activation('relu')(encoder)
    encoder = layers.Conv3D(num_filters, (3, 3, 3), padding='same')(encoder)
    encoder = layers.BatchNormalization()(encoder)
    encoder = layers.Activation('relu')(encoder)
    return encoder


def encoder_block(input_tensor, num_filters):
    encoder = conv_block(input_tensor, num_filters)
    encoder_pool = layers.MaxPooling3D((2, 2, 2), strides=(2, 2, 2))(encoder)

    return encoder_pool, encoder


def decoder_block(input_tensor, concat_tensor, num_filters):
    decoder = layers.Conv3DTranspose(
        num_filters, (2, 2, 2), strides=(2, 2, 2), padding='same')(input_tensor)
    decoder = layers.concatenate([concat_tensor, decoder], axis=-1)
    decoder = layers.BatchNormalization()(decoder)
    decoder = layers.Activation('relu')(decoder)
    decoder = layers.Conv3D(num_filters, (3, 3, 3), padding='same')(decoder)
    decoder = layers.BatchNormalization()(decoder)
    decoder = layers.Activation('relu')(decoder)
    decoder = layers.Conv3D(num_filters, (3, 3, 3), padding='same')(decoder)
    decoder = layers.BatchNormalization()(decoder)
    decoder = layers.Activation('relu')(decoder)
    return decoder


def dice_coeff(y_true, y_pred):
    smooth = 1.
    # Flatten
    y_true_f = tf.reshape(y_true, [-1])
    y_pred_f = tf.reshape(y_pred, [-1])
    intersection = tf.reduce_sum(y_true_f * y_pred_f)
    score = (2. * intersection + smooth) / (tf.reduce_sum(y_true_f) + tf.reduce_sum(y_pred_f) + smooth)
    return score


def dice_loss(y_true, y_pred):
    loss = 1 - dice_coeff(y_true, y_pred)
    return loss


def bce_dice_loss(y_true, y_pred):
    loss = losses.binary_crossentropy(y_true, y_pred) + dice_loss(y_true, y_pred)
    return loss


def bulid_model(weight_path):
    inputs = layers.Input(shape=(96, 96, 96, 1))
    # 256

    encoder0_pool, encoder0 = encoder_block(inputs, 32)
    # 128

    encoder1_pool, encoder1 = encoder_block(encoder0_pool, 64)
    # 64

    encoder2_pool, encoder2 = encoder_block(encoder1_pool, 128)
    # 32

    encoder3_pool, encoder3 = encoder_block(encoder2_pool, 256)
    # 16

    encoder4_pool, encoder4 = encoder_block(encoder3_pool, 512)
    # 8

    center = conv_block(encoder4_pool, 1024)
    # center

    decoder4 = decoder_block(center, encoder4, 512)
    # 16

    decoder3 = decoder_block(decoder4, encoder3, 256)
    # 32

    decoder2 = decoder_block(decoder3, encoder2, 128)
    # 64

    decoder1 = decoder_block(decoder2, encoder1, 64)
    # 128

    decoder0 = decoder_block(decoder1, encoder0, 32)
    # 256

    outputs = layers.Conv3D(1, (1, 1, 1), activation=None)(decoder0)

    model = models.Model(inputs=[inputs], outputs=[outputs])
    model.compile(optimizer='adam', loss=bce_dice_loss, metrics=[dice_loss])
    model.load_weights(weight_path)
    return model


def get_center(img):
    _img = img.copy()
    _img[img < 0.5] = 1
    _img[img > 0.5] = 0
    x, y, z = np.where(_img == 1)
    x = np.sum(x) / len(x)
    y = np.sum(y) / len(z)
    z = np.sum(z) / len(z)
    return np.array([z, y, x])


def get_radius(img):
    _img = img.copy()
    _img[img < 0.5] = 1
    _img[img > 0.5] = 0
    x, y, z = np.where(_img == 1.0)
    x.sort()
    y.sort()
    z.sort()
    l1 = x[0] - x[-1] - 1
    l2 = y[0] - y[-1] - 1
    l3 = z[0] - z[-1] - 1
    return np.abs(l1 + l2 + l3) / 6


'''

from acetabular_segment import bulid_model
model = bulid_model(acetabular_segmentation_model_2.h5)
pre = np.squeeze(model.predict(img[np.newaxis,:,:,:,np.newaxis]))

'''



import numpy as np
import pydicom
import vtk
from PySide2.QtGui import *
from PySide2.QtWidgets import *
from numba import jit

import vmi


def read_dicom():
    dcmdir = vmi.askDirectory()  # 用户选择文件夹

    if dcmdir is not None:  # 判断用户选中了有效文件夹并点击了确认
        series_list = vmi.sortSeries(dcmdir)  # 将文件夹及其子目录包含的所有DICOM文件分类到各个系列

        if len(series_list) > 0:  # 判断该文件夹内包含有效的DICOM系列
            series = vmi.askSeries(series_list)  # 用户选择DICOM系列

            if series is not None:  # 判断用户选中了有效系列并点击了确认
                with pydicom.dcmread(series.filenames()[0]) as ds:
                    global patient_name
                    patient_name = ds.PatientName
                return series.read()  # 读取DICOM系列为图像数据


def on_init_voi():
    size = vmi.askInt(1, 200, 1000, suffix='mm', title='请输入目标区域尺寸')
    if size is not None:
        global original_target, voi_size, voi_center
        original_target = original_view.pickPt_Cell()
        voi_center = np.array([vmi.imCenter(original_image)[0], original_target[1], original_target[2]])
        voi_size = np.array([vmi.imSize(original_image)[0], size, size])
        init_voi()

        voi_view.setCamera_Coronal()
        voi_view.setCamera_FitAll()
        voi_view.setEnabled(True)


@jit(nopython=True)
def ary_pad(input_ary, constant):
    for k in [0, input_ary.shape[0] - 1]:
        for j in range(input_ary.shape[1]):
            for i in range(input_ary.shape[2]):
                input_ary[k][j][i] = constant
    for k in range(input_ary.shape[0]):
        for j in [0, input_ary.shape[1] - 1]:
            for i in range(input_ary.shape[2]):
                input_ary[k][j][i] = constant
    for k in range(input_ary.shape[0]):
        for j in range(input_ary.shape[1]):
            for i in [0, input_ary.shape[2] - 1]:
                input_ary[k][j][i] = constant


def init_voi():
    global voi_size, voi_center, voi_origin, voi_cs
    voi_origin = voi_center - 0.5 * voi_size
    voi_cs = vmi.CS4x4(origin=voi_origin)

    global voi_image
    voi_image = vmi.imReslice(original_image, voi_cs, voi_size, -1024)
    voi_image.SetOrigin(*list(voi_cs.origin()))

    voi_ary = vmi.imArray_VTK(voi_image)
    ary_pad(voi_ary, -1024)
    voi_image = vmi.imVTK_Array(voi_ary, vmi.imOrigin(voi_image), vmi.imSpacing(voi_image))

    voi_volume.setData(voi_image)


def update_line_cut():
    if len(local_roi_pts) > 1:
        plcut_prop.setData(vmi.ccWire(vmi.ccSegments(local_roi_pts, True)))
    else:
        plcut_prop.setData()


@jit(nopython=True)
def ary_mask(input_ary, mask_ary, threshold, target):
    for k in range(input_ary.shape[0]):
        for j in range(input_ary.shape[1]):
            for i in range(input_ary.shape[2]):
                if mask_ary[k][j][i] != 0 and input_ary[k][j][i] >= threshold:
                    input_ary[k][j][i] = target


def plcut_voi_image():
    global voi_image

    cs = voi_view.cameraCS()
    d = vmi.imSize_Vt(voi_image, cs.axis(2))
    pts = [vmi.ptOnPlane(pt, vmi.imCenter(voi_image), cs.axis(2)) for pt in local_roi_pts]
    pts = [pt - d * cs.axis(2) for pt in pts]
    sh = vmi.ccFace(vmi.ccWire(vmi.ccSegments(pts, True)))
    sh = vmi.ccPrism(sh, cs.axis(2), 2 * d)

    mask = vmi.imStencil_PolyData(vmi.ccPd_Sh(sh),
                                  vmi.imOrigin(voi_image),
                                  vmi.imSpacing(voi_image),
                                  vmi.imExtent(voi_image))
    mask_ary = vmi.imArray_VTK(mask)
    voi_ary = vmi.imArray_VTK(voi_image)
    ary_mask(voi_ary, mask_ary, threshold, back_value)
    voi_image = vmi.imVTK_Array(voi_ary, vmi.imOrigin(voi_image), vmi.imSpacing(voi_image))
    voi_volume.setData(voi_image)


def update_cline_curve():
    if len(cline_centers) > 1:
        for i in range(len(cline_centers)):
            cline_center_props[i].setData(vmi.pdSphere(1, cline_centers[i]))

        edge = vmi.ccEdge(vmi.ccBSpline_Kochanek(cline_centers))
        l = vmi.ccLength(edge)
        pts = vmi.ccUniformPts_Edge(edge, round(max(l / 0.1, 2)))
        cline_curve_prop.setData(vmi.pdPolyline(pts))
    elif len(cline_centers) > 0:
        cline_center_props[0].setData(vmi.pdSphere(1, cline_centers[0]))
        cline_curve_prop.setData()
    else:
        cline_curve_prop.setData()
    voi_view.updateInTime()


def update_cline_wire():
    if cline_curve_prop.data().GetNumberOfPoints() >= 2:
        kd = vtk.vtkKdTreePointLocator()
        kd.SetDataSet(cline_curve_prop.data())
        kd.BuildLocator()

        for i in range(len(cline_centers)):
            ids = vtk.vtkIdList()
            kd.FindClosestNPoints(2, cline_centers[i], ids)

            pts = [np.array(kd.GetDataSet().GetPoint(ids.GetId(0))),
                   np.array(kd.GetDataSet().GetPoint(ids.GetId(1)))]
            n = pts[1] - pts[0]

            circle = vmi.ccCircle(cline_radiuss[i], cline_centers[i], n)
            cline_wires[i] = vmi.ccWire(circle)
            cline_wire_props[i].setData(cline_wires[i])


def update_cline_tube():
    if len(cline_wires) > 1:
        sh0 = vmi.ccFace(cline_wires[0])
        sh1 = vmi.ccLoft(cline_wires, redirect=True)
        sh2 = vmi.ccFace(cline_wires[-1])
        sh = vmi.ccSolid(vmi.ccSew([sh0, sh1, sh2]))
        cline_tube_prop.setData(sh)
    elif len(cline_wires) > 0:
        sh = vmi.ccFace(cline_wires[0])
        cline_tube_prop.setData(sh)
    else:
        cline_tube_prop.setData()


def LeftButtonPress(**kwargs):
    if kwargs['picked'] is voi_view:
        if local_rect.text() == '请勾画切割范围':
            pt = voi_view.pickPt_FocalPlane()
            local_roi_pts.clear()
            local_roi_pts.append(pt)
            update_line_cut()


def LeftButtonPressMove(**kwargs):
    if kwargs['picked'] is voi_view:
        if local_rect.text() == '请勾画切割范围':
            pt = voi_view.pickPt_Cell()

            for path_pt in local_roi_pts:
                if (path_pt == pt).all():
                    return

            local_roi_pts.append(pt)  # 添加拾取路径点
            update_line_cut()
        else:
            voi_view.mouseRotateFocal(**kwargs)
    elif kwargs['picked'] is original_slice:
        original_slice.windowMove(**kwargs)
        voi_view.updateInTime()
    elif kwargs['picked'] in cline_center_props:
        i = cline_center_props.index(kwargs['picked'])
        if not voi_volume.visible() and voi_slice.visible():
            pt = voi_view.pickPt_Cell()
            pt = vmi.rayPlane(voi_slice.slicePlaneOrigin(), voi_slice.slicePlaneNormal(),
                              pt, voi_view.camera('look'))
            if np.linalg.norm(cline_centers[i] - pt) < 10:
                cline_centers[i] = pt
        else:
            cline_centers[i] += voi_view.pickVt_FocalPlane()
        update_cline_curve()
        update_cline_wire()


def LeftButtonPressMoveRelease(**kwargs):
    if kwargs['picked'] is voi_view:
        if local_rect.text() == '请勾画切割范围':
            local_rect.draw_text('请耐心等待')
            plcut_voi_image()
            local_rect.draw_text('线切割')

            local_roi_pts.clear()
            update_line_cut()


def LeftButtonPressRelease(**kwargs):
    global cline_centers, cline_radiuss
    if kwargs['picked'] in [voi_view, voi_slice]:
        if kwargs['double'] is True:
            pt = voi_view.pickPt_Cell()

            cline_center_prop = vmi.PolyActor(voi_view, color=[0.6, 1, 0.6], pickable=True)
            cline_center_prop.setAlwaysOnTop(True)

            cline_center_prop.mouse['LeftButton']['PressMove'] = LeftButtonPressMove
            cline_center_prop.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease
            cline_center_prop.mouse['NoButton']['Wheel'] = NoButtonWheel

            cline_wire_prop = vmi.PolyActor(voi_view, color=[0.6, 1, 0.6], line_width=3, pickable=True)

            if len(cline_centers) > 1:
                if np.linalg.norm(pt - cline_centers[0]) < np.linalg.norm(pt - cline_centers[-1]):
                    cline_centers.insert(0, pt)
                    cline_radiuss.insert(0, 1)
                    cline_wires.insert(0, vmi.TopoDS_Shape())
                    cline_center_props.insert(0, cline_center_prop)
                    cline_wire_props.insert(0, cline_wire_prop)
                else:
                    cline_centers.append(pt)
                    cline_radiuss.append(1)
                    cline_wires.append(vmi.TopoDS_Shape())
                    cline_center_props.append(cline_center_prop)
                    cline_wire_props.append(cline_wire_prop)
            else:
                cline_centers.append(pt)
                cline_radiuss.append(1)
                cline_wires.append(vmi.TopoDS_Shape())
                cline_center_props.append(cline_center_prop)
                cline_wire_props.append(cline_wire_prop)
            update_cline_curve()
            update_cline_wire()
    elif kwargs['picked'] is voi_slice_visible_box:
        visible = voi_slice.visibleToggle()
        voi_slice_visible_box.draw_text(back_color=QColor('whitesmoke') if visible else QColor('black'))
    elif kwargs['picked'] is voi_volume_visible_box:
        visible = voi_volume.visibleToggle()
        voi_volume_visible_box.draw_text(back_color=QColor.fromRgbF(1, 1, 0.6) if visible else QColor('black'))
    elif kwargs['picked'] is threshold_box:
        global threshold
        v = vmi.askInt(-1000, threshold, 3000)
        if v is not None:
            threshold = v
            threshold_box.draw_text('阈值：{} HU'.format(threshold))
            voi_volume.setOpacityScalar({threshold - 1: 0, threshold: 1})
            voi_volume.setColor({threshold: [1, 0.4, 0.4], threshold + 100: [1, 1, 0.6]})
    elif kwargs['picked'] is local_rect:
        if local_rect.text() == '线切割':
            local_rect.draw_text('请勾画切割范围')
        elif local_rect.text() == '请勾画切割范围':
            local_rect.draw_text('线切割')
    elif kwargs['picked'] in cline_center_props:
        if kwargs['double'] is True:
            i = cline_center_props.index(kwargs['picked'])
            cline_center_props[i].delete()
            cline_wire_props[i].delete()
            del cline_centers[i], cline_radiuss[i], cline_wires[i]
            del cline_center_props[i], cline_wire_props[i]
            update_cline_curve()
            update_cline_wire()
        else:
            i = cline_center_props.index(kwargs['picked'])
            voi_view.camera().SetFocalPoint(*cline_centers[i])

            if cline_curve_prop.data().GetNumberOfPoints() >= 2:
                kd = vtk.vtkKdTreePointLocator()
                kd.SetDataSet(cline_curve_prop.data())
                kd.BuildLocator()

                ids = vtk.vtkIdList()
                kd.FindClosestNPoints(2, cline_centers[i], ids)

                pts = [np.array(kd.GetDataSet().GetPoint(ids.GetId(0))),
                       np.array(kd.GetDataSet().GetPoint(ids.GetId(1)))]
                n = pts[1] - pts[0]
                voi_slice.slicePlane(n, cline_centers[i])
            else:
                voi_slice.slicePlane(origin=cline_centers[i])
    elif kwargs['picked'] is cline_tube_box:
        cline_tube_box.draw_text('请耐心等待')
        update_cline_tube()
        cline_tube_box.draw_text('血管')
    elif kwargs['picked'] is cline_tube_visible_box:
        visible = cline_tube_prop.visibleToggle()
        cline_tube_visible_box.draw_text(back_color=QColor.fromRgbF(0.4, 0.6, 1) if visible else QColor('black'))
    elif kwargs['picked'] is cline_clear_box:
        for i in range(len(cline_centers)):
            cline_center_props[-1].delete()
            cline_wire_props[-1].delete()
            del cline_centers[-1], cline_radiuss[-1], cline_wires[-1]
            del cline_center_props[-1], cline_wire_props[-1]
        update_cline_curve()
        update_cline_wire()
    elif kwargs['picked'] is cline_data_open_box:
        file = vmi.askOpenFile(nameFilter='*.npz', title='npz')
        if file is not None:
            with open(file, mode='rb') as f:
                f.seek(0)
                data = np.load(f)
                print(data, sorted(data))
                if 'cline_centers' in data and 'cline_radius' in data:
                    cline_centers = list(data['cline_centers'])
                    cline_radiuss = list(data['cline_radius'])

                    for i in range(len(cline_centers)):
                        cline_wires.append(vmi.TopoDS_Shape())

                        prop = vmi.PolyActor(voi_view, color=[0.6, 1, 0.6], pickable=True)
                        prop.setAlwaysOnTop(True)
                        prop.mouse['LeftButton']['PressMove'] = LeftButtonPressMove
                        prop.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease
                        prop.mouse['NoButton']['Wheel'] = NoButtonWheel
                        cline_center_props.append(prop)

                        prop = vmi.PolyActor(voi_view, color=[0.6, 1, 0.6], line_width=3, pickable=True)
                        cline_wire_props.append(prop)

                    update_cline_curve()
                    update_cline_wire()
    elif kwargs['picked'] is cline_data_save_box:
        if len(cline_centers) > 0:
            file = vmi.askSaveFile(nameFilter='*.npz', title='npz')
            if file is not None:
                np.savez(file, cline_centers=np.array(cline_centers),
                         cline_radius=np.array(cline_radiuss))
    elif kwargs['picked'] is cline_tube_stl_box:
        if cline_tube_prop.data().GetNumberOfCells() > 0:
            file = vmi.askSaveFile(nameFilter='*.stl', title='stl')
            if file is not None:
                vmi.pdSaveFile_STL(cline_tube_prop.data(), file)


def NoButtonWheel(**kwargs):
    if kwargs['picked'] is threshold_box:
        global threshold
        threshold = min(max(threshold + 10 * kwargs['delta'], -1000), 3000)
        threshold_box.draw_text('阈值：{} HU'.format(threshold))
        voi_volume.setOpacityScalar({threshold - 1: 0, threshold: 1})
        voi_volume.setColor({threshold: [1, 0.4, 0.4], threshold + 100: [1, 1, 0.6]})
    elif kwargs['picked'] in cline_center_props:
        i = cline_center_props.index(kwargs['picked'])
        cline_radiuss[i] = min(max(cline_radiuss[i] + 0.1 * kwargs['delta'], 1), 10)
        update_cline_wire()


class Main(QScrollArea):
    def __init__(self):
        QScrollArea.__init__(self)
        brush = QBrush(QColor(255, 255, 255, 255))
        palette = QPalette()
        palette.setBrush(QPalette.Active, QPalette.Base, brush)
        palette.setBrush(QPalette.Active, QPalette.Window, brush)
        palette.setBrush(QPalette.Inactive, QPalette.Base, brush)
        palette.setBrush(QPalette.Inactive, QPalette.Window, brush)
        palette.setBrush(QPalette.Disabled, QPalette.Base, brush)
        palette.setBrush(QPalette.Disabled, QPalette.Window, brush)
        self.setPalette(palette)
        self.setWidgetResizable(True)

    def closeEvent(self, ev: QCloseEvent):
        ev.accept() if vmi.askYesNo('退出程序？') else ev.ignore()


if __name__ == '__main__':
    patient_name = str()
    original_image = read_dicom()  # 读取DICOM路径
    if original_image is None:
        vmi.appexit()

    # 0 原始图像
    original_target = np.zeros(3)
    original_view = vmi.View()

    original_slice = vmi.ImageSlice(original_view)  # 断层显示
    original_slice.setData(original_image)  # 载入读取到的DICOM图像
    original_slice.setColorWindow_Soft()
    original_slice.slicePlane(face='coronal', origin='center')  # 设置断层图像显示横断面，位置居中
    original_slice.mouse['LeftButton']['PressMove'] = LeftButtonPressMove

    original_slice_menu = vmi.Menu()
    original_slice_menu.menu.addAction('定位目标区域').triggered.connect(on_init_voi)
    original_slice_menu.menu.addSeparator()
    original_slice_menu.menu.addMenu(original_slice.menu)
    original_slice.mouse['RightButton']['PressRelease'] = original_slice_menu.menuexec

    patient_name_box = vmi.TextBox(original_view, text='姓名：{}'.format(patient_name),
                                   size=[0.24, 0.04], pos=[0, 0.04], anchor=[0, 0])

    original_view.setCamera_Coronal()
    original_view.setCamera_FitAll()  # 自动调整视图的视野范围

    # 1 目标区域
    voi_view = vmi.View()
    voi_view.mouse['LeftButton']['Press'] = LeftButtonPress
    voi_view.mouse['LeftButton']['PressMove'] = LeftButtonPressMove
    voi_view.mouse['LeftButton']['PressMoveRelease'] = LeftButtonPressMoveRelease
    voi_view.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease

    threshold, back_value = 120, -1100

    threshold_box = vmi.TextBox(voi_view, text='阈值：{} HU'.format(threshold),
                                size=[0.24, 0.04], pos=[0, 0.04], anchor=[0, 0], pickable=True)
    threshold_box.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease
    threshold_box.mouse['NoButton']['Wheel'] = NoButtonWheel

    voi_image = vtk.vtkImageData()
    voi_volume = vmi.ImageVolume(voi_view)
    voi_volume.setOpacityScalar({threshold - 1: 0, threshold: 1})
    voi_volume.setColor({threshold: [1, 0.4, 0.4], threshold + 100: [1, 1, 0.6]})

    voi_slice = vmi.ImageSlice(voi_view)
    voi_slice.setSlicePlaneNormal_Axial()
    voi_slice.visible(False)
    voi_slice.bindData(voi_volume)
    voi_slice.bindLookupTable(original_slice)
    voi_slice.mouse['LeftButton']['PressMove'] = voi_view.mouseRotateFocal

    voi_slice_visible_box = vmi.TextBox(voi_view, back_color=QColor('black'),
                                        size=[0.01, 0.04], pos=[0.24, 0.04], anchor=[1, 0], pickable=True)
    voi_slice_visible_box.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease

    voi_volume_visible_box = vmi.TextBox(voi_view, back_color=QColor.fromRgbF(1, 1, 0.6),
                                         size=[0.01, 0.04], pos=[0.23, 0.04], anchor=[1, 0], pickable=True)
    voi_volume_visible_box.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease

    voi_size, voi_center, voi_origin, voi_cs = np.zeros(3), np.zeros(3), np.zeros(3), vmi.CS4x4()

    # 线切割
    local_roi_pts = []
    plcut_prop = vmi.PolyActor(voi_view, color=[1, 0.4, 0.4], line_width=3)
    plcut_prop.setAlwaysOnTop(True)

    local_rect = vmi.TextBox(voi_view, text='线切割', pickable=True,
                             size=[0.24, 0.04], pos=[0, 0.1], anchor=[0, 0])
    local_rect.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease

    # 中心线
    cline_centers, cline_center_props = [], []
    cline_radiuss, cline_wires, cline_wire_props = [], [], []
    cline_curve_prop = vmi.PolyActor(voi_view, color=[0.4, 0.6, 1], line_width=3)
    cline_tube_prop = vmi.PolyActor(voi_view, color=[0.4, 0.6, 1])

    cline_tube_box = vmi.TextBox(voi_view, text='血管', pickable=True,
                                 size=[0.24, 0.04], pos=[0, 0.16], anchor=[0, 0])
    cline_tube_box.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease

    cline_tube_visible_box = vmi.TextBox(voi_view, back_color=QColor.fromRgbF(0.4, 0.6, 1),
                                         size=[0.01, 0.04], pos=[0.24, 0.16], anchor=[1, 0], pickable=True)
    cline_tube_visible_box.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease

    cline_clear_box = vmi.TextBox(voi_view, text='清空', pickable=True,
                                  size=[0.06, 0.04], pos=[0, 0.2], anchor=[0, 0])
    cline_clear_box.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease

    cline_data_open_box = vmi.TextBox(voi_view, text='导入', pickable=True,
                                      size=[0.06, 0.04], pos=[0.06, 0.2], anchor=[0, 0])
    cline_data_open_box.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease

    cline_data_save_box = vmi.TextBox(voi_view, text='导出', pickable=True,
                                      size=[0.06, 0.04], pos=[0.12, 0.2], anchor=[0, 0])
    cline_data_save_box.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease

    cline_tube_stl_box = vmi.TextBox(voi_view, text='STL', pickable=True,
                                     size=[0.06, 0.04], pos=[0.18, 0.2], anchor=[0, 0])
    cline_tube_stl_box.mouse['LeftButton']['PressRelease'] = LeftButtonPressRelease

    # 测量视图

    # 视图布局
    views = [original_view, voi_view]

    screen_width = QGuiApplication.screens()[0].geometry().width()
    for v in views:
        v.setMinimumWidth(0.5 * screen_width)

    for v in views[1:]:
        v.setEnabled(False)

    widget = QWidget()
    layout = QGridLayout(widget)
    layout.addWidget(original_view, 0, 0, 3, 1)
    layout.addWidget(voi_view, 0, 1, 3, 1)

    main = Main()
    main.setWidget(widget)

    vmi.appexec(main)  # 执行主窗口程序
    vmi.appexit()  # 清理并退出程序



import hashlib
import gridfs
import pymongo
import pickle
import vtk

import vmi

import SimpleITK


def from_image_to_stl():
    # 获取图像信息
    uid = "1.2.840.113704.1.111.9620.1487668799.1"
    search_dict = {'StudyInstanceUID': uid}
    md5_key = image_data_db.find_one(search_dict)
    data_image = image_file_db.find_one({'md5': md5_key['predicted_md5'][-1]})

    pelvis_image = pickle.loads(data_image.read())

    # 中值滤波
    res = 1
    spc = vmi.imSpacing(pelvis_image)
    size = [int(0.5 + res / spc[0]),
            int(0.5 + res / spc[1]),
            int(0.5 + res / spc[2])]

    itk_image: SimpleITK.Image = vmi.imSITK_VTK(pelvis_image)
    smooth = SimpleITK.BinaryMedianImageFilter()
    smooth.SetRadius(size)
    itk_image = smooth.Execute(itk_image)
    pelvis_image = vmi.imVTK_SITK(itk_image)

    # 等值面提取
    fe = vtk.vtkFlyingEdges3D()
    fe.SetInputData(pelvis_image)
    fe.SetValue(0, 1)
    fe.SetComputeNormals(0)
    fe.SetComputeGradients(0)
    fe.SetComputeScalars(0)
    fe.Update()

    # 光滑处理
    smooth = vtk.vtkWindowedSincPolyDataFilter()
    smooth.SetInputData(fe.GetOutput())
    smooth.SetNumberOfIterations(20)
    smooth.SetBoundarySmoothing(0)
    smooth.SetFeatureEdgeSmoothing(0)
    smooth.SetFeatureAngle(60)
    smooth.SetPassBand(0.1)
    smooth.SetNonManifoldSmoothing(1)
    smooth.SetNormalizeCoordinates(1)
    smooth.Update()

    pd = smooth.GetOutput()

    # 保存STL文件
    stl_filename = uid + '.stl'
    w = vtk.vtkSTLWriter()
    w.SetInputData(pd)
    w.SetFileName(stl_filename)
    w.SetFileTypeToBinary()
    w.Update()

    # 上传STL文件
    stl_file = open(stl_filename, 'rb')
    file_data = stl_file.read()
    model_md5 = hashlib.md5(file_data).hexdigest()

    if not model_file_db.exists({'md5': model_md5}):
        file_id = model_file_db.put(file_data,
                                    fileType='.stl',
                                    StudyInstanceUID=uid,
                                    type='model')

        # 检查上传结果是否与本地一致
        if model_file_db.find_one({'_id': file_id}).md5 != model_md5:
            model_file_db.delete(file_id)
            return print('文件上传失败')
        else:
            # 替换旧数据,增加STL数据的md5
            if not model_data_db.find_one(search_dict):
                model_data_db.insert_one(search_dict)

            key_data_dict = model_data_db.find_one(search_dict)
            value_data_dict = {**key_data_dict,
                               'model_md5': model_md5,
                               'type': 'model'}
            model_data_db.find_one_and_replace(key_data_dict, value_data_dict)

    return pd


def from_cup_data_to_stl():
    pt_dict = {'臼杯中心': [107.93227321, 179.52647828, 157.30932222],
               '臼杯半径': 25.0,
               '朝向': [1, 0, 0]}

    o, r, n = pt_dict['臼杯中心'], pt_dict['臼杯半径'], pt_dict['朝向']

    plane = vtk.vtkPlane()
    plane.SetNormal(n)
    plane.SetOrigin(o)

    pd = vmi.pdSphere(r, o)
    pd = vmi.pdClip_Implicit(pd, plane)[1]
    pd_border = vmi.pdPolyline_Regular(r, o, n)
    pd = vmi.pdAppend([pd, pd_border])

    return pd


def return_globals():
    return globals()


if __name__ == '__main__':
    client = pymongo.MongoClient('mongodb://root:medraw123@192.168.11.122:27017/admin', 27017)

    image_data_db: pymongo.collection.Collection = client.xj_testDB.pelvisPredict
    image_file_db = gridfs.GridFS(client.xj_testDB, collection='pelvisPredict')

    model_data_db: pymongo.collection.Collection = client.xj_testDB.pelvisSTL
    model_file_db = gridfs.GridFS(client.xj_testDB, collection='pelvisSTL')

    view = vmi.View()

    # 骨盆模型
    pelvis_data = from_image_to_stl()
    pelvis_prop = vmi.PolyActor(view)
    pelvis_prop.setData(pelvis_data)

    # 臼杯
    cup_data = from_cup_data_to_stl()
    cup_prop = vmi.PolyActor(view)
    cup_prop.setData(cup_data)

    view.setCamera_FitAll()

    main = vmi.Main(return_globals)
    main.setAppName('image_to_stl')
    main.setAppVersion(vmi.version)
    main.excludeKeys += ['main']

    main.layout().addWidget(view, 0, 0, 1, 1)
    vmi.appexec(main)  # 执行主窗口程序
    vmi.appexit()  # 清理并退出程序



import hashlib
import json
import pathlib
import pickle
import tempfile
import threading
from typing import Dict, Any

import gridfs
import pydicom
import pymongo
import vtk

import vmi
from PySide2.QtGui import *
from PySide2.QtWidgets import *
import numpy as np
import nibabel as nib
import os
from copy import deepcopy


def local_open_dicom():
    global series_list
    dcmdir = vmi.askDirectory('DICOM')

    if dcmdir is not None:
        series_list = vmi.sortSeries(dcmdir)
        local_select_series()


def local_select_series():
    global series_dict
    series = vmi.askSeries(series_list)
    if series is not None:
        series_dict = series.toKeywordValue()
        input_image(series.read())


def input_image(image: vtk.vtkImageData):
    series_volume.setData(image)
    series_view.setCamera_FitAll()
    series_view.updateAtOnce()
    vmi.appupdate()


def update_series_opcity():
    series_volume.setOpacityScalar({series_opacity_range[0] - 1: 0,
                                    series_opacity_range[0]: 1,
                                    series_opacity_range[1]: 1,
                                    series_opacity_range[1] + 1: 0})
    series_volume.setColor({series_opacity_range[0]: [1, 1, 0.9],
                            series_opacity_range[1]: [1, 1, 0.1]})


def local_save_nifti():
    vmi.imSaveFile_NIFTI(series_volume.data())


def upload(image: vtk.vtkImageData, kw_value: Dict[str, Any]):
    # 删除已有数据
    uid = {'SeriesInstanceUID': kw_value['SeriesInstanceUID']}

    for one in file_db.find(uid):
        file_db.delete(one._id)

    for one in data_db.find(uid):
        data_db.delete_one(one)

    # 图像序列化为文件，计算特征码，上传文件项
    with tempfile.TemporaryDirectory() as p:
        p = pathlib.Path(p) / '.nii'
        vmi.imSaveFile_NIFTI(image, str(p))
        image_bytes = p.read_bytes()
        file_id = file_db.put(image_bytes,
                              fileType='.nii',
                              scalarType=image.GetScalarTypeAsString(),
                              SeriesInstanceUID=kw_value['SeriesInstanceUID'])

    # 检查上传结果是否与本地一致
    if file_db.find_one({'_id': file_id}).md5 != hashlib.md5(image_bytes).hexdigest():
        file_db.delete(file_id)
        return print('远端MD5校验失败 {}'.format(repr(kw_value)))

    # 上传数据项
    data_db.insert_one(kw_value)


def _dialog_ok_cancel(widgets, title, ok_cancel=True) -> QDialog:
    dialog = QDialog()
    dialog.setSizePolicy(SizePolicyMinimum)
    dialog.setWindowTitle(title)
    dialog.setLayout(QVBoxLayout())
    dialog.setMinimumWidth(800)

    for w in widgets:
        w.setParent(dialog)
        dialog.layout().addWidget(w)

    if ok_cancel:
        button = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, dialog)
        button.setSizePolicy(SizePolicyMinimum)
        button.accepted.connect(dialog.accept)

        button.rejected.connect(dialog.reject)
        dialog.layout().addWidget(button)
    return dialog


def download_by_SeriesInstanceUID():
    client = pymongo.MongoClient('mongodb://root:medraw123@192.168.11.122:27017/admin', 27017)
    file_path = vmi.askOpenFile(nameFilter='*.npy', title='请选择SeriesInstanceUID所在的文件')
    collection_list = client.testDB.list_collection_names()
    collection_list = sorted(collection_list)
    tree = QTreeWidget()
    tree.setColumnCount(2)
    tree.setHeaderLabels(['序号', '数据库'])
    for ith, col in enumerate(collection_list):
        item = QTreeWidgetItem()
        item.setText(0, str(ith))
        item.setText(1, str(col))
        tree.addTopLevelItem(item)

    d = _dialog_ok_cancel([tree], "选择.file数据库")
    if d.exec_() == QDialog.Accepted:
        collection_name = tree.currentItem().text(1).split('.')[0]
        file_db = gridfs.GridFS(client.testDB, collection=collection_name)
    save_path = vmi.askDirectory()
    SeriesInstanceUID_list = np.load(file_path)
    for uid in SeriesInstanceUID_list:
        search_dict = {'SeriesInstanceUID': uid}
        find_data = file_db.find_one(search_dict)
        image_data = pickle.loads(find_data.read())
        np_data = vmi.imArray_VTK(image_data)
        np_data = np.flipud(np_data.swapaxes(0, 2))
        affine = np.zeros((4, 4))
        origin = vmi.imOrigin(image_data)
        spacing = vmi.imSpacing(image_data)
        for i, j in enumerate(spacing):
            affine[i][i] = j
            affine[i][-1] = origin[i]
        affine[-1][-1] = 1.0
        nib_data = nib.Nifti1Image(np_data, affine)
        nib_file_name = os.path.join(save_path, str(uid) + '.nii.gz')
        nib.save(nib_data, nib_file_name)


def upload_nii_gz():
    client = pymongo.MongoClient('mongodb://root:medraw123@192.168.11.122:27017/admin', 27017)
    file_names = vmi.askOpenFiles('*.nii.gz', '请选择nii.gz文件')
    unmarkData_nii_db: pymongo.collection.Collection = client.testDB.spinePelvis_nii_gz
    unmarkFile_nii_db = gridfs.GridFS(client.testDB, collection='spinePelvis_nii_gz')
    for i, p in enumerate(file_names):
        file_path = os.path.split(p)[-1]
        f = open(p, 'rb')
        md5 = hashlib.md5(f).hexdigest()
        print(md5)
        uid = file_path.split('.nii.gz')[0]
        # 禁止上传相同SeriesInstanceUID的nii.gz
        if not unmarkData_nii_db.find_one({'SeriesInstanceUID': uid}):
            unmarkData_nii_db.insert_one({'SeriesInstanceUID': uid, 'nii_gz_md5': [md5], 'file_name': file_path})
            unmarkFile_nii_db.put(f.read(), **{'fileType': '.nii.gz', 'SeriesInstanceUID': uid, 'file_name': file_path})
        f.close()


def upload_marked_nii_gz():
    client = pymongo.MongoClient('mongodb://root:medraw123@192.168.11.122:27017/admin', 27017)
    file_names = vmi.askOpenFiles('*.nii.gz', '请选择已标注nii.gz文件')
    markData_nii_db: pymongo.collection.Collection = client.testDB.spinePelvis_Marked_nii_gz
    markFile_nii_db = gridfs.GridFS(client.testDB, collection='spinePelvis_Marked_nii_gz')
    for i, p in enumerate(file_names):
        file_path = os.path.split(p)[-1]
        f = open(p, 'rb').read()
        md5 = hashlib.md5(f).hexdigest()
        print(md5)

        uid = file_path.split('.nii.gz')[0]
        # 上传多版本标注，以list存储MD5
        if not markData_nii_db.find_one({'SeriesInstanceUID': uid}):
            markData_nii_db.insert_one({'SeriesInstanceUID': uid, 'nii_gz_md5': [md5], 'file_name': file_path})
            markFile_nii_db.put(f, **{'fileType': '.nii.gz', 'SeriesInstanceUID': uid, 'file_name': file_path})

        else:

            marked_data_dict = markData_nii_db.find_one({'SeriesInstanceUID': uid})
            marked_data_dict_new = deepcopy(marked_data_dict)
            marked_data_dict_new['nii_gz_md5'].append(md5)
            markData_nii_db.find_one_and_replace(marked_data_dict, marked_data_dict_new)

            markFile_nii_db.put(f, **{'fileType': '.nii.gz', 'SeriesInstanceUID': uid, 'file_name': file_path})


def download_nii_gz():
    client = pymongo.MongoClient('mongodb://root:medraw123@192.168.11.122:27017/admin', 27017)
    collection_list = client.testDB.list_collection_names()
    collection_list = sorted(collection_list)
    tree = QTreeWidget()
    tree.setColumnCount(2)
    tree.setHeaderLabels(['序号', '数据库'])
    for ith, col in enumerate(collection_list):
        item = QTreeWidgetItem()
        item.setText(0, str(ith))
        item.setText(1, str(col))
        tree.addTopLevelItem(item)

    d = _dialog_ok_cancel([tree], "选择.file数据库")
    if d.exec_() == QDialog.Accepted:
        collection_name = tree.currentItem().text(1).split('.')[0]
        file_db = gridfs.GridFS(client.testDB, collection=collection_name)
    tree.setHeaderLabels(['序号', 'SeriesInstanceUID'])
    file_tree = QTreeWidget()
    file_tree.setColumnCount(3)
    file_tree.setHeaderLabels(['序号', 'SeriesInstanceUID', 'md5'])
    for i, file in enumerate(file_db.find({"SeriesInstanceUID": {'$exists': True}})):
        print(file._file)
        item = QTreeWidgetItem()
        item.setText(0, str(i + 1))
        item.setText(1, str(file._file["SeriesInstanceUID"]))
        item.setText(2, str(file._file["md5"]))

        file_tree.addTopLevelItem(item)
    d = _dialog_ok_cancel([file_tree], "选择下载数据")
    if d.exec_() == QDialog.Accepted:
        save_path = vmi.askDirectory('请选择保存位置')
        uid = file_tree.currentItem().text(1)
        find_file = file_db.find_one({"SeriesInstanceUID": uid})
        save_file_name = find_file._file['file_name']
        full_save_path = os.path.join(save_path, save_file_name)
        s = open(full_save_path, 'wb')
        s.write(find_file.read())
        s.close()


def remote_auto_upload():
    global data_db, file_db
    client = pymongo.MongoClient('mongodb://root:medraw123@192.168.11.122:27017/admin', 27017)
    data_db = client.testDB.get_collection('spinePelvis')
    file_db = gridfs.GridFS(client.testDB, collection='spinePelvis')

    # 远端清空
    data_db.delete_many({})
    for i in file_db.find():
        file_db.delete(i._id)
    chunks_db: pymongo.collection.Collection = client.testDB.get_collection('spinePelvis.chunks')
    chunks_db.delete_many({})

    p = vmi.askDirectory('DICOM根目录')
    if p is None:
        return

    # 本地排序
    save_json = pathlib.Path(p) / 'filenames_list.json'
    if save_json.exists():
        filenames_list = json.loads(save_json.read_text())
    else:
        filenames_list = vmi.sortFilenames(p)
        pathlib.Path(str(save_json)).write_text(json.dumps(filenames_list))

    # 本地过滤
    save_json = pathlib.Path(p) / 'filenames_list_1.json'
    if save_json.exists():
        filenames_list = json.loads(save_json.read_text())
    else:
        new_list, i = {}, 0
        for filenames in filenames_list:
            i = i + 1

            ds = pydicom.dcmread(filenames[0])

            if 'PatientID' not in ds or 'StudyInstanceUID' not in ds or 'SeriesInstanceUID' not in ds:
                continue
            if 'Modality' not in ds or ds['Modality'].value != 'CT':
                continue
            if 'BitsStored' not in ds or ds['BitsStored'].value != 12:
                continue
            if 'StudyDescription' in ds and 'Spine' in ds['StudyDescription'].value:
                continue
            if 'SliceThickness' not in ds or ds['SliceThickness'].value == '' or float(ds['SliceThickness'].value) > 1:
                continue
            if 'SliceLocation' not in ds or len(filenames) < 2:
                continue

            kw = str(ds['StudyInstanceUID'].value)
            if kw in new_list and len(filenames) < len(new_list[kw]):
                continue

            ds1 = pydicom.dcmread(filenames[1])
            delta = abs(float(ds['SliceLocation'].value) - float(ds1['SliceLocation'].value))
            if delta != float(ds['SliceThickness'].value):
                continue

            new_list[kw] = filenames.copy()
            print('本地过滤 {}/{}'.format(len(new_list), i))
        filenames_list = new_list.copy()
        pathlib.Path(str(save_json)).write_text(json.dumps(filenames_list))

    # 图像过滤
    save_json = pathlib.Path(p) / 'filenames_list_2.json'
    if save_json.exists():
        filenames_list = json.loads(save_json.read_text())
    else:
        new_list = []
        for kw in filenames_list:
            filenames = filenames_list[kw]

            series = vmi.SeriesData(filenames)
            image = series.read()

            tags = series.toKeywordValue()
            t = threading.Thread(target=upload, args=[image, tags])
            t.start()
            vmi.appwait(t)

            input_image(image)
            save_snapshot = pathlib.Path(p) / kw
            series_view.snapshot_PNG(str(save_snapshot))

            new_list.append(tags['SeriesInstanceUID'])
            print('上传成功 {}/{}'.format(len(new_list), len(filenames_list)))
        filenames_list = new_list.copy()
        pathlib.Path(str(save_json)).write_text(json.dumps(filenames_list))


def LeftButtonPressRelease(**kwargs):
    global series_opacity_range
    if kwargs['picked'] is series_opacity_range_box[0]:
        v = vmi.askInt(-10000, series_opacity_range[0], series_opacity_range[1])
        if v is not None:
            series_opacity_range[0] = v
        series_opacity_range_box[0].draw_text('min {:.0f}'.format(series_opacity_range[0]))
        update_series_opcity()
    elif kwargs['picked'] is series_opacity_range_box[1]:
        v = vmi.askInt(series_opacity_range[0], series_opacity_range[0], 10000)
        if v is not None:
            series_opacity_range[1] = v
        series_opacity_range_box[1].draw_text('min {:.0f}'.format(series_opacity_range[1]))
        update_series_opcity()


def NoButtonWheel(**kwargs):
    global series_opacity_range
    if kwargs['picked'] is series_opacity_range_box[0]:
        series_opacity_range[0] = min(max(series_opacity_range[0] + kwargs['delta'],
                                          -10000), series_opacity_range[1])
        series_opacity_range_box[0].draw_text('min {:.0f}'.format(series_opacity_range[0]))
        update_series_opcity()
    elif kwargs['picked'] is series_opacity_range_box[1]:
        series_opacity_range[1] = min(max(series_opacity_range[1] + kwargs['delta'],
                                          series_opacity_range[0]), 10000)
        series_opacity_range_box[1].draw_text('max {:.0f}'.format(series_opacity_range[1]))
        update_series_opcity()


def return_globals():
    return globals()


if __name__ == '__main__':
    main = vmi.Main(return_globals)
    main.setAppName('dicom_to_series')
    main.setAppVersion(vmi.version)
    main.excludeKeys += ['main']

    menu_local = main.menuBar().addMenu('本地')
    menu_local.addAction('打开DICOM').triggered.connect(local_open_dicom)
    menu_local.addAction('载入Series').triggered.connect(local_select_series)
    menu_local.addAction('保存NIFTI').triggered.connect(local_save_nifti)

    menu_remote = main.menuBar().addMenu('远端')
    # menu_remote.addAction('载入Series').triggered.connect(remote_get_series)
    # menu_remote.addAction('自动上传').triggered.connect(remote_auto_upload)
    menu_remote.addAction('按SeriesInstanceUID下载').triggered.connect(download_by_SeriesInstanceUID)
    menu_remote.addAction('上传原始nii.gz文件').triggered.connect(upload_nii_gz)
    menu_remote.addAction('上传标记nii.gz文件').triggered.connect(upload_marked_nii_gz)
    menu_remote.addAction('下载nii.gz文件').triggered.connect(download_nii_gz)

    series_list, series_dict = [], {}
    series_view = vmi.View()
    series_view.setCamera_Coronal()
    series_volume = vmi.ImageVolume(series_view)

    series_opacity_range = [200, 3000]
    update_series_opcity()
    SizePolicyMinimum = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

    series_opacity_range_box = [vmi.TextBox(series_view, text='min {:.0f}'.format(series_opacity_range[0]),
                                            size=[0.1, 0.03], pos=[0, 0.03], anchor=[0, 0], pickable=True),
                                vmi.TextBox(series_view, text='max {:.0f}'.format(series_opacity_range[1]),
                                            size=[0.1, 0.03], pos=[0.1, 0.03], anchor=[0, 0], pickable=True)]
    for box in series_opacity_range_box:
        box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]
        box.mouse['NoButton']['Wheel'] = [NoButtonWheel]

    main.layout().addWidget(series_view, 0, 0, 1, 1)
    vmi.appexec(main)
    vmi.appexit()



import pathlib
import tempfile
import time

import numpy as np
import pydicom
import vtk
from PySide2.QtGui import *
from numba import jit

import vmi
import vmi.application.skyward


def read_dicom():
    dcmdir = vmi.askDirectory('DICOM')  # 用户选择文件夹

    if dcmdir is not None:  # 判断用户选中了有效文件夹并点击了确认
        series_list = vmi.sortSeries(dcmdir)  # 将文件夹及其子目录包含的所有DICOM文件分类到各个系列

        if len(series_list) > 0:  # 判断该文件夹内包含有效的DICOM系列
            global series
            series = vmi.askSeries(series_list)  # 用户选择DICOM系列

            if series is not None:  # 判断用户选中了有效系列并点击了确认
                with pydicom.dcmread(series.filenames()[0]) as ds:
                    global patient_name
                    patient_name = str(ds.PatientName)
                    patient_name_box.draw_text('姓名：{}'.format(patient_name))
                    patient_name_box.setVisible(True)

                    global dicom_ds
                    dicom_ds = series.toKeywordValue()
                return series.read()  # 读取DICOM系列为图像数据


def on_init_voi():
    size = vmi.askInt(1, 300, 1000, suffix='mm', title='请输入目标区域尺寸')
    if size is not None:
        global LR, original_target, voi_size, voi_center, spcut_center
        original_target = original_view.pickPt_Cell()
        voi_center = np.array([vmi.imCenter(original_slice.data())[0], original_target[1], original_target[2]])
        voi_size = np.array([vmi.imSize(original_slice.data())[0], size, size])
        init_voi()

        LR = 1 if original_target[0] > voi_center[0] else -1
        patient_LR_box.draw_text('患侧：{}'.format('左' if LR > 0 else '右'))
        patient_LR_box.setVisible(True)

        spcut_center = original_target.copy()
        update_spcut()

        voi_view.setCamera_Coronal()
        update_spcut_section()

        voi_view.setCamera_FitAll()
        for v in [voi_view, *spcut_y_views, *spcut_z_views]:
            v.setEnabled(True)


@jit(nopython=True)
def ary_pad(input_ary, constant):
    for k in [0, input_ary.shape[0] - 1]:
        for j in range(input_ary.shape[1]):
            for i in range(input_ary.shape[2]):
                input_ary[k][j][i] = constant
    for k in range(input_ary.shape[0]):
        for j in [0, input_ary.shape[1] - 1]:
            for i in range(input_ary.shape[2]):
                input_ary[k][j][i] = constant
    for k in range(input_ary.shape[0]):
        for j in range(input_ary.shape[1]):
            for i in [0, input_ary.shape[2] - 1]:
                input_ary[k][j][i] = constant


def init_voi():
    global voi_size, voi_center, voi_origin, voi_cs, voi_image
    voi_origin = voi_center - 0.5 * voi_size
    voi_cs = vmi.CS4x4(origin=voi_origin)

    voi_image = vmi.imReslice(original_slice.data(), voi_cs, voi_size, -1024)
    voi_image.SetOrigin(voi_cs.origin())

    voi_ary = vmi.imArray_VTK(voi_image)
    ary_pad(voi_ary, -1024)
    voi_image = vmi.imVTK_Array(voi_ary, vmi.imOrigin(voi_image), vmi.imSpacing(voi_image))
    voi_volume.setData(voi_image)

    pd = vmi.imIsosurface(voi_image, 3000)
    metal_prop.setData(pd)
    pelvis_view.setCamera_Coronal()
    pelvis_view.setCamera_FitAll()
    clear_guide_plate()


def update_line_cut():
    if len(plcut_pts) > 1:
        plcut_prop.setData(vmi.ccWire(vmi.ccSegments(plcut_pts, True)))
    else:
        plcut_prop.setData(None)
    clear_guide_plate()


@jit(nopython=True)
def ary_mask(input_ary, mask_ary, threshold, target):
    for k in range(input_ary.shape[0]):
        for j in range(input_ary.shape[1]):
            for i in range(input_ary.shape[2]):
                if mask_ary[k][j][i] != 0 and threshold <= input_ary[k][j][i]:
                    input_ary[k][j][i] = target


def plcut_voi_image():
    cs = voi_view.cameraCS()
    d = vmi.imSize_Vt(voi_volume.data(), cs.axis(2))
    pts = [vmi.ptOnPlane(pt, vmi.imCenter(voi_volume.data()), cs.axis(2)) for pt in plcut_pts]
    pts = [pt - d * cs.axis(2) for pt in pts]
    sh = vmi.ccFace(vmi.ccWire(vmi.ccSegments(pts, True)))
    sh = vmi.ccPrism(sh, cs.axis(2), 2 * d)

    mask = vmi.imStencil_PolyData(vmi.ccPd_Sh(sh),
                                  vmi.imOrigin(voi_volume.data()),
                                  vmi.imSpacing(voi_volume.data()),
                                  vmi.imExtent(voi_volume.data()))
    mask_ary = vmi.imArray_VTK(mask)
    voi_ary = vmi.imArray_VTK(voi_volume.data())
    ary_mask(voi_ary, mask_ary, bone_value, target_value)
    voi_image = vmi.imVTK_Array(voi_ary, vmi.imOrigin(voi_volume.data()), vmi.imSpacing(voi_volume.data()))
    voi_volume.setData(voi_image)

    for v in spcut_z_views + spcut_y_views:
        v.updateInTime()
    clear_guide_plate()


def spcut_voi_image():
    spcut_image = vmi.imStencil_PolyData(spcut_prop.data(),
                                         vmi.imOrigin(voi_volume.data()),
                                         vmi.imSpacing(voi_volume.data()),
                                         vmi.imExtent(voi_volume.data()))
    spcut_ary = vmi.imArray_VTK(spcut_image)
    voi_ary = vmi.imArray_VTK(voi_volume.data())

    ary_mask(voi_ary, spcut_ary, bone_value, target_value)
    voi_image = vmi.imVTK_Array(voi_ary, vmi.imOrigin(voi_volume.data()), vmi.imSpacing(voi_volume.data()))
    voi_volume.setData(voi_image)
    clear_guide_plate()


def update_spcut():
    spcut_mesh = vmi.pdSphere(spcut_radius, spcut_center)
    spcut_prop.setData(spcut_mesh)
    clear_guide_plate()


def update_spcut_section():
    angle = 30
    y_oblique = [vmi.CS4x4().rotate(angle, [0, 0, 1]), vmi.CS4x4(), vmi.CS4x4().rotate(-angle, [0, 0, 1])]
    z_oblique = [vmi.CS4x4().rotate(-angle, [0, 1, 0]), vmi.CS4x4(), vmi.CS4x4().rotate(angle, [0, 1, 0])]

    y_normals = [cs.mvt([0, 1, 0]) for cs in y_oblique]
    z_normals = [cs.mvt([0, 0, 1]) for cs in z_oblique]

    y_cs = [vmi.CS4x4_Coronal(), vmi.CS4x4_Coronal(), vmi.CS4x4_Coronal()]
    y_cs[0] = y_cs[0].rotate(-angle, y_cs[0].axis(1))
    y_cs[2] = y_cs[2].rotate(angle, y_cs[2].axis(1))

    z_cs = [vmi.CS4x4_Axial(), vmi.CS4x4_Axial(), vmi.CS4x4_Axial()]
    z_cs[0] = z_cs[0].rotate(-angle, z_cs[0].axis(1))
    z_cs[2] = z_cs[2].rotate(angle, z_cs[2].axis(1))

    hheight = 50
    hwidth = hheight * spcut_y_views[0].width() / spcut_y_views[0].height()

    for i in range(3):
        y_cs[i].setOrigin(spcut_center - y_cs[i].mpt([hwidth, hheight, 0]))
        z_cs[i].setOrigin(spcut_center - z_cs[i].mpt([hwidth, hheight, 0]))

    for i in range(3):
        spcut_y_slices[i].setSlicePlane(spcut_center, y_normals[i])
        spcut_z_slices[i].setSlicePlane(spcut_center, z_normals[i])

    for i in range(3):
        spcut_y_props[i].setData(vmi.pdPolygon_Regular(spcut_radius, spcut_center, y_normals[i]))
        spcut_z_props[i].setData(vmi.pdPolygon_Regular(spcut_radius, spcut_center, z_normals[i]))

    for i in range(3):
        spcut_y_views[i].setCamera_FPlane(y_cs[i], 2 * hheight)
        spcut_z_views[i].setCamera_FPlane(z_cs[i], 2 * hheight)
    clear_guide_plate()


def init_pelvis():
    pd = vmi.imResample_Isotropic(voi_volume.data())
    pd = vmi.imIsosurface(pd, bone_value)
    pelvis_prop.setData(pd)

    pelvis_view.setCamera_Coronal()
    pelvis_view.setCamera_FitAll()
    pelvis_view.setEnabled(True)

    origin = np.array(pelvis_prop.data().GetCenter())
    pelvis_cs.setOrigin(origin)
    cup_cs.setOrigin(original_target)
    clear_guide_plate()


def update_pelvis_MPP():
    global pelvis_cs
    vr = pelvis_view.cameraVt_Right()
    vr = vr if np.dot(vr, np.array([1, 0, 0])) >= 0 else -vr
    if vmi.vtAngle(vr, pelvis_cs.axis(1)) < 1 or vmi.vtAngle(vr, pelvis_cs.axis(1)) > 179:
        return

    pelvis_cs.setAxis(0, vr)
    pelvis_cs.orthogonalize([2, 0])
    update_cup_orientation()
    update_cup()

    origin = pelvis_view.pickPt_FocalPlane([0.5 * pelvis_view.width(), 0.5 * pelvis_view.height()])
    pelvis_cs.setOrigin(origin)
    cs = vmi.CS4x4().reflect(pelvis_cs.axis(0), pelvis_cs.origin())
    pd = vmi.pdMatrix(pelvis_prop.data(), cs.matrix4x4())
    pelvis_MPP_prop.setData(pd)

    cs = pelvis_view.cameraCS()
    pd = vmi.pdMatrix(pelvis_prop.data(), cs.inv().matrix4x4())
    b, c = pd.GetBounds(), pd.GetCenter()
    left_top = np.array([b[0], b[2], c[2]])
    right_top = np.array([b[1], b[2], c[2]])
    left_bottom = np.array([b[0], b[3], c[2]])
    right_bottom = np.array([b[1], b[3], c[2]])
    xn = int(np.round(0.1 * (pd.GetBounds()[1] - pd.GetBounds()[0])))

    lines = []
    for i in range(xn + 1):
        pts = [left_top + i / xn * (left_bottom - left_top),
               right_top + i / xn * (right_bottom - right_top)]
        pts = [cs.mpt(pt) for pt in pts]
        line = vmi.ccEdge(vmi.ccSegment(pts[0], pts[1]))
        line = vmi.ccPd_Sh(line)
        lines.append(line)

    pd = vmi.pdAppend(lines)
    pelvis_cs_indicator_prop.setData(pd)
    clear_guide_plate()


def update_pelvis_APP():
    global pelvis_cs
    vr = pelvis_view.cameraVt_Right()
    vr = vr if np.dot(vr, np.array([0, 1, 0])) >= 0 else -vr
    if vmi.vtAngle(vr, pelvis_cs.axis(0)) < 1 or vmi.vtAngle(vr, pelvis_cs.axis(0)) > 179:
        return

    pelvis_cs.setAxis(1, vr)
    pelvis_cs.orthogonalize([2, 0])
    update_cup_orientation()
    update_cup()

    cs = pelvis_view.cameraCS()
    pd = vmi.pdMatrix(pelvis_prop.data(), cs.inv().matrix4x4())
    b, c = pd.GetBounds(), pd.GetCenter()
    left_top = np.array([b[0], b[2], c[2]])
    right_top = np.array([b[1], b[2], c[2]])
    left_bottom = np.array([b[0], b[3], c[2]])
    right_bottom = np.array([b[1], b[3], c[2]])
    xn = int(np.round(0.1 * (pd.GetBounds()[1] - pd.GetBounds()[0])))

    lines = []
    for i in range(xn + 1):
        pts = [left_top + i / xn * (right_top - left_top),
               left_bottom + i / xn * (right_bottom - left_bottom)]
        pts = [cs.mpt(pt) for pt in pts]
        line = vmi.ccEdge(vmi.ccSegment(pts[0], pts[1]))
        line = vmi.ccPd_Sh(line)
        lines.append(line)

    pd = vmi.pdAppend(lines)
    pelvis_cs_indicator_prop.setData(pd)
    clear_guide_plate()


def update_cup_orientation():
    cs = pelvis_cs.rotate(-cup_RI, pelvis_cs.axis(1)).rotate(-cup_RA, pelvis_cs.axis(0))
    cup_axis = -cs.axis(2) * np.array([LR, 1, 1])

    cup_cs.setAxis(1, pelvis_cs.axis(1))
    cup_cs.setAxis(2, cup_axis)
    cup_cs.orthogonalize([0, 1])
    cup_cs.setAxis(0, -LR * cup_cs.axis(0))
    cup_cs.normalize()
    clear_guide_plate()

    # cup_RA = np.rad2deg(np.arctan(-cup_axis[1] / np.sqrt(1 - cup_axis[2] ** 2)))
    # cup_RI = np.rad2deg(-np.abs(cup_axis[0] / cup_axis[2]))


def update_cup():
    plane = vtk.vtkPlane()
    plane.SetNormal(cup_cs.axis(2))
    plane.SetOrigin(cup_cs.origin())

    pd = vmi.pdSphere(cup_radius, cup_cs.origin())
    pd = vmi.pdClip_Implicit(pd, plane)[1]

    pd_border = vmi.pdPolyline_Regular(cup_radius, cup_cs.origin(), cup_cs.axis(2))
    pd = vmi.pdAppend([pd, pd_border])

    cs = vmi.CS4x4().reflect(pelvis_cs.axis(0), pelvis_cs.origin())
    pd_MPP = vmi.pdMatrix(pd, cs.matrix4x4())
    cup_prop.setData(vmi.pdAppend([pd, pd_MPP]))

    if pelvis_slice_box.text() == '断层':
        pd = vmi.pdCut_Implicit(cup_prop.data(), pelvis_slice.slicePlane())
        cup_section_prop.setData(pd)
    clear_guide_plate()


@jit(nopython=True)
def pt_ary_cs_mask(pt_ary: np.ndarray, cs: np.ndarray, mask: np.ndarray):
    return (pt_ary @ cs) * mask


def update_cup_camera_sagittal():
    cs = vmi.CS4x4(axis1=-pelvis_cs.axis(2), axis2=-LR * pelvis_cs.axis(0))
    cs.orthogonalize([0])
    cs.setOrigin(cup_cs.origin() - 2.25 * cup_radius * cs.axis(0) - 2.25 * cup_radius * cs.axis(1))
    pelvis_view.setCamera_FPlane(cs, 4.5 * cup_radius)
    pelvis_view.updateInTime()


def update_cup_camera_coronal():
    cs = vmi.CS4x4(axis1=-pelvis_cs.axis(2), axis2=pelvis_cs.axis(1))
    cs.orthogonalize([0])
    cs.setOrigin(cup_cs.origin() - 2.25 * cup_radius * cs.axis(0) - 2.25 * cup_radius * cs.axis(1))
    pelvis_view.setCamera_FPlane(cs, 4.5 * cup_radius)
    pelvis_view.updateInTime()


def update_cup_camera_axial():
    cs = vmi.CS4x4(axis1=pelvis_cs.axis(1), axis2=pelvis_cs.axis(2))
    cs.orthogonalize([0])
    cs.setOrigin(cup_cs.origin() - 2.25 * cup_radius * cs.axis(0) - 2.25 * cup_radius * cs.axis(1))
    pelvis_view.setCamera_FPlane(cs, 4.5 * cup_radius)
    pelvis_view.updateInTime()


def update_cup_camera_RA():
    cs = vmi.CS4x4(axis1=-cup_cs.axis(1), axis2=-cup_cs.axis(0))
    cs.orthogonalize([0])
    cs.setOrigin(cup_cs.origin() - 2.25 * cup_radius * cs.axis(0) - 2.25 * cup_radius * cs.axis(1))
    pelvis_view.setCamera_FPlane(cs, 4.5 * cup_radius)
    pelvis_view.updateInTime()


def update_cup_camera_RI():
    cs = vmi.CS4x4(axis1=-cup_cs.axis(0), axis2=cup_cs.axis(1))
    cs.orthogonalize([0])
    cs.setOrigin(cup_cs.origin() - 2.25 * cup_radius * cs.axis(0) - 2.25 * cup_radius * cs.axis(1))
    pelvis_view.setCamera_FPlane(cs, 4.5 * cup_radius)
    pelvis_view.updateInTime()


def update_cup_camera_plane():
    cs = vmi.CS4x4(axis1=-cup_cs.axis(0), axis2=-cup_cs.axis(2))
    cs.orthogonalize([0])
    cs.setOrigin(cup_cs.origin() - 2.25 * cup_radius * cs.axis(0) - 2.25 * cup_radius * cs.axis(1))
    pelvis_view.setCamera_FPlane(cs, 4.5 * cup_radius)
    pelvis_view.updateInTime()


def init_guide():
    size = np.array([5 * cup_radius, 5 * cup_radius, 5 * cup_radius])
    cs = vmi.CS4x4(origin=cup_cs.origin() - 0.5 * size)
    image = vmi.imReslice(voi_volume.data(), cs, size, -1024)
    image.SetOrigin(cs.origin())

    ary = vmi.imArray_VTK(image)
    ary_pad(ary, -1024)
    image = vmi.imVTK_Array(ary, vmi.imOrigin(image), vmi.imSpacing(image))
    image = vmi.imResample_Isotropic(image)

    pd = vmi.imIsosurface(image, bone_value)
    pd = vmi.pdSmooth_WindowedSinc(pd, 20)
    aceta_prop.setData(pd)

    update_guide_path()
    update_cup_camera_plane()
    guide_view.setEnabled(True)


def update_guide_path():
    global cup_cs, body_cs, large_z, small_t
    large_z, small_t = 100, 5
    body_vt_x = cup_cs.rotate(LR * (body_angle + 45), -cup_cs.axis(2)).mvt([1, 0, 0])
    body_cs = vmi.CS4x4(origin=cup_cs.mpt([0, 0, large_z]), axis0=body_vt_x, axis2=-cup_cs.axis(2))
    body_cs.orthogonalize([1, 0])
    body_cs.normalize()

    global body_xr, body_yr, head_xr, head_yr, body_x_negtive
    head_xr = 7  # 头部长径，臼缘匹配块
    head_yr = body_yr
    body_x_negtive = body_yr / np.sin(np.deg2rad(72))

    global body_wire
    body_pts = [body_cs.mpt([body_xr + head_xr, LR * -body_yr, 0]),
                body_cs.mpt([body_xr + head_xr, LR * body_yr, 0]),
                body_cs.mpt([-body_x_negtive * np.cos(np.deg2rad(72)), LR * body_yr, 0]),
                body_cs.mpt([-body_x_negtive, 0, 0]),
                body_cs.mpt([-body_x_negtive * np.cos(np.deg2rad(72)), LR * -body_yr, 0])]

    body_wire = body_cs.workplane().polyline(
        [body_cs.inv().mpt(pt)[:2].tolist() for pt in body_pts]).close().val().wrapped
    body_axis = vmi.ccEdge(vmi.ccSegment(body_cs.mpt([0, 0, 0]), body_cs.mpt([0, 0, 2 * large_z])))

    global body_hole_wire
    body_hole_radius = small_t
    body_hole_wire = vmi.ccWire(vmi.ccEdge(
        vmi.ccCircle(body_hole_radius, body_cs.origin(), body_cs.axis(2), body_cs.axis(0))))

    global head_cs
    head_cs = body_cs.copy()
    head_cs.setOrigin(body_cs.mpt([body_xr, 0, 0]))

    global arm_length, arm_area_wire, shaft_radius
    shaft_radius = 5
    arm_area_pts = [head_cs.mpt([-arm_length * np.sin(np.deg2rad(t)),
                                    -LR * (arm_length * np.cos(np.deg2rad(t))), 0]) for t in range(-30, 60)]
    arm_area_wire = vmi.ccWire(vmi.ccSegments(arm_area_pts))

    global guide_path, guide_path_prop
    guide_path = vmi.ccBoolean_Union([body_wire, body_axis, body_hole_wire, arm_area_wire])
    guide_path_prop.setData(guide_path)
    return


def update_guide_plate():
    global UDI, UDI_time
    UDI_time = time.localtime()
    UDI = '{}{}'.format('TH', vmi.convert_base(round(time.time() * 100)).upper())

    time_start, time_prev = time.time(), time.time()

    global kirs_radius, kirs_hole_radius
    kirs_hole_radius = kirs_radius + 0.2

    global body_cs, body_xr, body_yr, head_cs, head_xr, head_yr, large_z, small_t, arm_length, shaft_radius
    mold_xr = body_xr + head_xr
    mold_yr = arm_length + kirs_hole_radius + 2
    res = 1
    xn, yn = int(np.ceil(mold_xr / res)) + 1, int(np.ceil(mold_yr / res)) + 1
    body_mold, body_z_array = [], []
    body = vmi.ccMatchSolid_Rect(
        aceta_prop.data(), body_cs, xn, yn, res, small_t, 2 * large_z - small_t, 0.2, body_mold, body_z_array)

    time_stop = time.time() - time_prev
    time_prev = time.time()
    print('match body', time_stop)

    nx, ny = int(np.ceil(head_xr / res)) + 1, int(np.ceil(head_yr / res)) + 1
    head_z_array = []
    vmi.ccMatchSolid_Rect(
        aceta_prop.data(), head_cs, nx, ny, res, small_t, 2 * large_z - small_t, 0.2, None, head_z_array)

    time_stop = time.time() - time_prev
    time_prev = time.time()
    print('match head', time_stop)

    global head_height
    head_height = 15
    head_z = np.min(head_z_array[0]) - small_t

    global psi_mold, psi_mold_prop
    mold_max_z = head_z + small_t + body_xr + head_xr
    mold_wire = body_cs.workplane([0, -LR * 0.5 * (mold_yr - mold_xr), 0]).rect(
        2 * mold_xr, mold_xr + mold_yr).val().wrapped
    mold_box = vmi.ccPrism(vmi.ccFace(mold_wire), body_cs.axis(2), mold_max_z)
    psi_mold = vmi.ccBoolean_Intersection([body_mold[0], mold_box])

    time_stop = time.time() - time_prev
    time_prev = time.time()
    print('psi_mold', time_stop)

    label_cs = vmi.CS4x4(origin=body_cs.mpt([0, LR * 0.5 * (mold_xr - mold_yr), mold_max_z - 1]),
                         axis1=-LR * body_cs.axis(1), axis2=-body_cs.axis(2))
    label_cs.orthogonalize([0])

    label_wire = label_cs.workplane().rect(2 * mold_xr, mold_xr + mold_yr + small_t * 4).val().wrapped
    label_box = vmi.ccPrism(vmi.ccFace(label_wire), body_cs.axis(2), 2)
    psi_mold = vmi.ccBoolean_Union([psi_mold, label_box])

    label_cs = vmi.CS4x4(origin=body_cs.mpt([0, -LR * (mold_yr + small_t), mold_max_z]),
                         axis1=-LR * body_cs.axis(1), axis2=-body_cs.axis(2))
    label_cs.orthogonalize([0])
    text = '{} {}'.format(patient_name, 'L' if LR > 0 else 'R')
    label_text = label_cs.workplane().text(text, small_t, 3, font='DengXian', kind='bold').val().wrapped
    psi_mold = vmi.ccBoolean_Union([psi_mold, label_text])

    label_cs = vmi.CS4x4(origin=body_cs.mpt([0, LR * (mold_xr + small_t), mold_max_z]),
                         axis1=-LR * body_cs.axis(1), axis2=-body_cs.axis(2))
    label_cs.orthogonalize([0])
    label_text = label_cs.workplane().text(UDI, small_t, 3, font='DengXian', kind='bold').val().wrapped
    psi_mold = vmi.ccBoolean_Union([psi_mold, label_text])

    time_stop = time.time() - time_prev
    time_prev = time.time()
    print('label', time_stop)

    global body_x_negtive
    body_pts = [[body_xr + head_xr, LR * -body_yr, head_z],
                [body_xr + head_xr, LR * body_yr, head_z],
                [-body_x_negtive * np.cos(np.deg2rad(72)), LR * body_yr, head_z],
                [-body_x_negtive, 0, head_z],
                [-body_x_negtive * np.cos(np.deg2rad(72)), LR * -body_yr, head_z]]
    body_wire = vmi.ccWire(vmi.ccSegments([body_cs.mpt(pt) for pt in body_pts], True))
    body_box = vmi.ccPrism(vmi.ccFace(body_wire), body_cs.axis(2), 2 * large_z)
    body = vmi.ccBoolean_Intersection([body, body_box])

    if match_style == '下沉式':
        body_pts = [[body_xr + head_xr, LR * body_yr, 2 * large_z],
                    [body_xr + head_xr, LR * body_yr, head_z],
                    [body_xr - head_xr, LR * body_yr, head_z],
                    [0, LR * body_yr, head_z + (body_xr - head_xr) * np.tan(np.deg2rad(45))],
                    [-body_x_negtive, LR * body_yr, head_z + (body_xr - head_xr) * np.tan(np.deg2rad(45))],
                    [-body_x_negtive, LR * body_yr, 2 * large_z]]
        body_wire = vmi.ccWire(vmi.ccSegments([body_cs.mpt(pt) for pt in body_pts], True))
        body_box = vmi.ccPrism(vmi.ccFace(body_wire), -LR * body_cs.axis(1), 2 * body_yr)
        body = vmi.ccBoolean_Intersection([body, body_box])
    else:
        body_pts = [[body_xr + head_xr, LR * body_yr, large_z * 2],
                    [body_xr + head_xr, LR * body_yr, head_z],
                    [-body_x_negtive, LR * body_yr, head_z],
                    [-body_x_negtive, LR * body_yr, head_z + small_t],
                    [body_xr - head_xr * 2, LR * body_yr, head_z + small_t],
                    [body_xr - head_xr * 2, LR * body_yr, 2 * large_z]]
        body_wire = vmi.ccWire(vmi.ccSegments([body_cs.mpt(pt) for pt in body_pts], True))
        body_box = vmi.ccPrism(vmi.ccFace(body_wire), -LR * body_cs.axis(1), 2 * body_yr)
        body = vmi.ccBoolean_Intersection([body, body_box])

    global shaft
    h = head_height + 1 - 3
    dr = h * np.tan(np.deg2rad(5.675))
    shaft_wire0 = vmi.ccWire(vmi.ccEdge(vmi.ccCircle(
        shaft_radius, head_cs.mpt([0, 0, head_z + 1]), head_cs.axis(2), head_cs.axis(0))))
    shaft_wire1 = vmi.ccWire(vmi.ccEdge(vmi.ccCircle(
        shaft_radius - dr, head_cs.mpt([0, 0, head_z + 1 - h]), head_cs.axis(2), head_cs.axis(0))))
    shaft_face0 = vmi.ccFace(shaft_wire0)
    shaft_face1 = vmi.ccFace(shaft_wire1)
    shaft = vmi.ccSolid(vmi.ccSew([vmi.ccLoft([shaft_wire0, shaft_wire1]), shaft_face0, shaft_face1]))
    body = vmi.ccBoolean_Union([body, shaft])

    global shaft_hole
    shaft_hole_wire = vmi.ccWire(vmi.ccEdge(
        vmi.ccCircle(kirs_hole_radius, head_cs.mpt([0, 0, 0]), head_cs.axis(2), head_cs.axis(0))))
    shaft_hole = vmi.ccPrism(vmi.ccFace(shaft_hole_wire), head_cs.axis(2), 2 * large_z)
    body = vmi.ccBoolean_Difference([body, shaft_hole])

    time_stop = time.time() - time_prev
    time_prev = time.time()
    print('body', time_stop)

    global body_hole_wire
    body_hole = vmi.ccPrism(vmi.ccFace(body_hole_wire), body_cs.axis(2), 2 * large_z)

    global psi_arm
    arm_pts = [[kirs_hole_radius + 2, -LR * (arm_length + kirs_hole_radius + 2), head_z],
                  [kirs_hole_radius + 2, -LR * (arm_length - kirs_hole_radius - 2), head_z],
                  [head_xr, LR * -head_xr, head_z],
                  [head_xr, LR * head_xr, head_z],
                  [-head_xr, LR * head_xr, head_z],
                  [-head_xr, LR * -head_xr, head_z],
                  [-kirs_hole_radius - 2, -LR * (arm_length - kirs_hole_radius - 2), head_z],
                  [-kirs_hole_radius - 2, -LR * (arm_length + kirs_hole_radius + 2), head_z]]
    arm_pts = [head_cs.mpt(pt) for pt in arm_pts]
    arm_wire = vmi.ccWire(vmi.ccSegments(arm_pts, True))
    psi_arm = vmi.ccPrism(vmi.ccFace(arm_wire), -head_cs.axis(2), head_height + 2)

    h = head_height + 1 - 2
    dr = h * np.tan(np.deg2rad(5.675))
    hole_wire0 = vmi.ccWire(vmi.ccEdge(vmi.ccCircle(shaft_radius, head_cs.mpt([0, 0, head_z + 1]),
                                                    head_cs.axis(2), head_cs.axis(0))))
    hole_wire1 = vmi.ccWire(vmi.ccEdge(vmi.ccCircle(shaft_radius - dr, head_cs.mpt([0, 0, head_z + 1 - h]),
                                                    head_cs.axis(2), head_cs.axis(0))))
    hole_face0 = vmi.ccFace(hole_wire0)
    hole_face1 = vmi.ccFace(hole_wire1)
    hole = vmi.ccSolid(vmi.ccSew([vmi.ccLoft([hole_wire0, hole_wire1]), hole_face0, hole_face1]))
    psi_arm = vmi.ccBoolean_Difference([psi_arm, hole])

    arm_hole_center = head_cs.mpt([0, -LR * arm_length, 0])
    arm_hole = vmi.ccPrism(vmi.ccFace(vmi.ccWire(vmi.ccCircle(
        kirs_hole_radius, arm_hole_center, head_cs.axis(2), head_cs.axis(0)))), head_cs.axis(2), 2 * large_z)
    psi_arm = vmi.ccBoolean_Difference([psi_arm, arm_hole])
    psi_arm = vmi.ccBoolean_Difference([psi_arm, shaft_hole])

    psi_arm_prop.setData(psi_arm)
    # psi_arm_prop.setData(vmi.pdTriangle(vmi.ccPd_Sh(psi_arm), 0, 0))

    center_hole_wire = vmi.ccWire(vmi.ccEdge(
        vmi.ccCircle(kirs_hole_radius, body_cs.mpt([0, 0, 0]), body_cs.axis(2), body_cs.axis(0))))
    center_hole = vmi.ccPrism(vmi.ccFace(center_hole_wire), body_cs.axis(2), 2 * large_z)

    psi_mold = vmi.ccBoolean_Difference([psi_mold, center_hole])
    psi_mold = vmi.ccBoolean_Difference([psi_mold, shaft_hole])
    psi_mold_prop.setData(psi_mold)
    # mold_prop.setData(vmi.pdTriangle(vmi.ccPd_Sh(mold), 0, 0))

    time_stop = time.time() - time_prev
    time_prev = time.time()
    print('arm', time_stop)

    time_stop = time.time() - time_prev
    time_prev = time.time()
    print('arm', time_stop)

    global psi_match, psi_match_prop
    psi_match = vmi.ccBoolean_Difference([body, body_hole])
    psi_match_prop.setData(psi_match)
    # psi_match_prop.setData(vmi.pdTriangle(vmi.ccPd_Sh(psi_match), 0, 0))

    time_stop = time.time() - time_prev
    time_prev = time.time()
    print('body cut', time_stop)
    print(time.time() - time_start)


def clear_guide_plate():
    global psi_match, psi_arm, psi_mold
    psi_match = vmi.TopoDS_Shape()
    psi_arm = vmi.TopoDS_Shape()
    psi_mold = vmi.TopoDS_Shape()
    psi_match_prop.setData(vtk.vtkPolyData())
    psi_arm_prop.setData(vtk.vtkPolyData())
    psi_mold_prop.setData(vtk.vtkPolyData())


def case_sync():
    global cup_radius, cup_RA, cup_RI
    with tempfile.TemporaryDirectory() as p:
        case_vti = {'voi_vti': voi_volume.data()}
        for kw in case_vti:
            f = str(pathlib.Path(p) / '{}.vtp'.format(kw))
            vmi.imSaveFile_XML(case_vti[kw], f)
            case_vti[kw] = f

        case_vtp = {'pelvis_vtp': pelvis_prop.data(),
                    'psi_match_vtp': psi_match_prop.data(),
                    'psi_arm_vtp': psi_arm_prop.data(),
                    'psi_mold_vtp': psi_mold_prop.data()}
        for kw in case_vtp:
            f = str(pathlib.Path(p) / '{}.vtp'.format(kw))
            vmi.pdSaveFile_XML(case_vtp[kw], f)
            case_vtp[kw] = f

        case_stl = {'psi_match_stl': psi_match_prop.data(),
                    'psi_arm_stl': psi_arm_prop.data(),
                    'psi_mold_stl': psi_mold_prop.data()}
        for kw in case_stl:
            f = str(pathlib.Path(p) / '{}.stl'.format(kw))
            vmi.pdSaveFile_STL(case_stl[kw], f)
            case_stl[kw] = f

        spc, dim = vmi.imSpacing(original_slice.data()), vmi.imDimensions(original_slice.data())
        pelvis_dicom_info = {'Modality': dicom_ds['Modality'],
                             'StudyDate': dicom_ds['StudyDate'],
                             'StudyDescription': dicom_ds['StudyDescription'],
                             'Resolution': dim.tolist(),
                             'Spacing': spc.tolist()}

        case = vmi.application.skyward.case(order_model='TH', order_uid=UDI,
                                            order_files=['psi_match_stl', 'psi_arm_stl', 'psi_mold_stl'],
                                            patient_name=dicom_ds['PatientName'],
                                            patient_sex=dicom_ds['PatientSex'],
                                            patient_age=dicom_ds['PatientAge'],
                                            dicom_files={'pelvis_dicom': series.filenames()},
                                            vti_files=case_vti,
                                            vtp_files=case_vtp,
                                            stl_files=case_stl,
                                            pelvis_dicom_info=pelvis_dicom_info,
                                            surgical_side='left' if LR > 0 else 'right',
                                            pelvis_cs=pelvis_cs.array4x4().tolist(),
                                            head_cs=head_cs.array4x4().tolist(),
                                            cup_cs=cup_cs.array4x4().tolist(),
                                            cup_radius=round(cup_radius * 100) / 100,
                                            cup_RA=round(cup_RA * 100) / 100,
                                            cup_RI=round(cup_RI * 100) / 100,
                                            bone_value=bone_value,
                                            body_angle=body_angle,
                                            body_xr=body_xr,
                                            body_yr=body_yr,
                                            arm_length=arm_length,
                                            match_style=match_style)

        case = vmi.application.skyward.case_sync(case, ['pelvis_dicom', 'order_model', 'surgical_side'])

        if case is True:
            vmi.askInfo('同步完成')
        elif case:
            cup_radius = case['cup_radius']
            cup_RA = case['cup_RA']
            cup_RI = case['cup_RI']
            cup_radius_box.draw_text('半径：{:.0f} mm'.format(cup_radius))
            cup_RA_box.draw_text('前倾角：{:.1f}°'.format(cup_RA))
            cup_RI_box.draw_text('外展角：{:.1f}°'.format(cup_RI))
            update_cup_orientation()
            update_cup()

            if case['customer_confirmed']:
                init_guide()
            vmi.askInfo('同步完成')


def save_docx(file):
    global UDI, UDI_time

    with tempfile.TemporaryDirectory() as p:
        png = str(pathlib.Path(p) / '.png')

        doc = vmi.docx_document('设计确认单')
        vmi.docx_section_landscape_A4(doc.sections[0])
        vmi.docx_section_margin_middle(doc.sections[0])
        vmi.docx_section_header(doc.sections[0])

        s_title = vmi.docx_style(doc, '影为标题', 'Normal', size=26, alignment='CENTER')
        s_header = vmi.docx_style(doc, '影为页眉', 'Header', size=10)
        s_body_left = vmi.docx_style(doc, '影为正文', 'Body Text')
        s_body_left_indent = vmi.docx_style(doc, '影为正文首行缩进', 'Body Text', first_line_indent=2)
        s_body_center = vmi.docx_style(doc, '影为正文居中', 'Body Text', alignment='CENTER')
        s_table = vmi.docx_style(doc, '影为表格', 'Table Grid', alignment='CENTER')

        doc.sections[0].header.paragraphs[0].text = '影为医疗科技（上海）有限公司\t\t\t\t{}'.format('[表单号]')
        doc.sections[0].header.paragraphs[0].style = s_header

        def add_row_text(table, texts, styles, merge=None):
            r = len(table.add_row().table.rows) - 1
            if merge is not None:
                table.cell(r, merge[0]).merge(table.cell(r, merge[1]))
            for c in range(min(len(table.columns), len(texts))):
                table.cell(r, c).paragraphs[0].text = texts[c]
                table.cell(r, c).paragraphs[0].style = styles[c]

        def add_row_picture(table, text, styles, pic_file):
            r = len(table.add_row().table.rows) - 1
            table.cell(r, 1).merge(table.cell(r, 3))
            table.cell(r, 0).paragraphs[0].text = text
            table.cell(r, 0).paragraphs[0].style = styles[0]
            table.cell(r, 1).paragraphs[0].style = styles[1]

            with open(pic_file, 'rb') as pic:
                table.cell(r, 1).paragraphs[0].add_run().add_picture(pic, width=table.cell(r, 1).width * 0.7)

        doc_table = doc.add_table(1, 4, s_table)
        doc_table.autofit = True
        doc_table.alignment = vmi.docx.enum.table.WD_TABLE_ALIGNMENT.CENTER

        doc_table.cell(0, 0).merge(doc_table.cell(0, 3))
        doc_table.cell(0, 0).paragraphs[0].text = '设计确认单\nDesign Confirmation'
        doc_table.cell(0, 0).paragraphs[0].style = s_title

        add_row_text(doc_table, ['顾客姓名\nCustomer name', '',
                                 '顾客编号\nCustomer UID', ''], [s_body_center] * 4)
        add_row_text(doc_table, ['单位\nOrganization', '',
                                 '电邮\nEmail', ''], [s_body_center] * 4)
        add_row_text(doc_table, ['邮寄信息\nMailing information', ''], [s_body_center, s_body_left], [1, 3])

        text = '顾客收货视为已阅读并接受：'
        text += '\n1）所述产品属于个性化医疗器械；'
        text += '\n2）确认设计输入已包含顾客要求；'
        text += '\n3）确认设计输出符合顾客要求，同意进行生产制造；'
        text += '\n4）当顾客要求与产品上市信息冲突时，不作为设计输入。'
        add_row_text(doc_table, ['销售条款\nTerms of sale', text], [s_body_center, s_body_left], [1, 3])

        add_row_text(doc_table, ['销售工程师\nSales engineer', '',
                                 '电话\nTelephone', ''], [s_body_center] * 4)

        add_row_text(doc_table, ['产品名称\nProduct name', '全髋关节置换手术导板',
                                 '序列号\nSerial number', UDI], [s_body_center] * 4)
        add_row_text(doc_table, ['型号\nModel', 'TH',
                                 '规格\nSpecification', '定制式'], [s_body_center] * 4)
        add_row_text(doc_table, [
            '患者\nPatient',
            '{}\n{} {}'.format(dicom_ds['PatientName'], dicom_ds['PatientSex'], dicom_ds['PatientAge']),
            '手术侧\nSurgical side', '左' if LR > 0 else '右'], [s_body_center] * 4)

        spc, dim = vmi.imSpacing(original_slice.data()), vmi.imDimensions(original_slice.data())
        text = '描述：{}，{}，{}'.format(dicom_ds['StudyDate'], dicom_ds['Modality'], dicom_ds['StudyDescription'])
        text += '\n参数：{}×{}×{}，{}×{}×{} mm'.format(dim[0], dim[1], dim[2], spc[0], spc[1], spc[2])
        text += '\n系列：{}'.format(dicom_ds['SeriesInstanceUID'])
        add_row_text(doc_table, ['影像输入\nRadiological input', text], [s_body_center, s_body_left], [1, 3])

        text = '影像学前倾角：{:.1f}°\n影像学外展角：{:.1f}°'.format(cup_RA, cup_RI)
        add_row_text(doc_table, ['临床输入\nClinical input', text], [s_body_center, s_body_left], [1, 3])

        text = '骨阈值：{:.0f} HU'.format(bone_value)
        text += '\n主体方位：{:.0f}°，主体长径：{:.0f} mm，主体短径：{:.0f} mm'.format(body_angle, body_xr, body_yr)
        text += '\n旋臂长度：{:.0f} mm'.format(arm_length)
        text += '\n匹配方式：{}'.format(match_style)
        add_row_text(doc_table, ['造型输入\nModeling input', text], [s_body_center, s_body_left], [1, 3])

        add_row_text(doc_table, ['设计时间\nDesign time', '{}年{}月{}日{}'.format(
            UDI_time.tm_year, UDI_time.tm_mon, UDI_time.tm_mday, time.strftime('%H:%M:%S', UDI_time))],
                     [s_body_center, s_body_left], [1, 3])

        text = '{} {}'.format(UDI, '定位板.stl')
        text += '\n{} {}'.format(UDI, '旋臂.stl')
        text += '\n{} {}'.format(UDI, '配模.stl')
        add_row_text(doc_table, ['设计输出\nDesign output', text], [s_body_center, s_body_left], [1, 3])

        visibles = [aceta_prop.visible(),
                    guide_path_prop.visible(),
                    psi_match_prop.visible(),
                    psi_arm_prop.visible(),
                    psi_mold_prop.visible()]

        aceta_prop.setVisible(True)
        guide_path_prop.setVisible(False)
        psi_match_prop.setVisible(False)
        psi_arm_prop.setVisible(False)
        psi_mold_prop.setVisible(False)
        update_cup_camera_plane()
        guide_view.snapshot_PNG(png)
        add_row_picture(doc_table, '快照（1/6）\nSnapshot', [s_body_center, s_body_center], png)

        aceta_prop.setVisible(True)
        guide_path_prop.setVisible(True)
        psi_match_prop.setVisible(True)
        psi_arm_prop.setVisible(False)
        psi_mold_prop.setVisible(False)
        guide_view.snapshot_PNG(png)
        add_row_picture(doc_table, '快照（2/6）\nSnapshot', [s_body_center, s_body_center], png)

        aceta_prop.setVisible(True)
        guide_path_prop.setVisible(False)
        psi_match_prop.setVisible(True)
        psi_arm_prop.setVisible(True)
        psi_mold_prop.setVisible(False)
        guide_view.camera().Elevation(-30)
        guide_view.camera().OrthogonalizeViewUp()
        guide_view.snapshot_PNG(png)
        add_row_picture(doc_table, '快照（3/6）\nSnapshot', [s_body_center, s_body_center], png)

        aceta_prop.setVisible(True)
        guide_path_prop.setVisible(False)
        psi_match_prop.setVisible(True)
        psi_arm_prop.setVisible(True)
        psi_mold_prop.setVisible(False)
        guide_view.camera().Elevation(-60)
        guide_view.camera().OrthogonalizeViewUp()
        guide_view.camera().Azimuth(LR * 90)
        guide_view.camera().Elevation(30)
        guide_view.camera().OrthogonalizeViewUp()
        guide_view.snapshot_PNG(png)
        add_row_picture(doc_table, '快照（4/6）\nSnapshot', [s_body_center, s_body_center], png)

        aceta_prop.setVisible(False)
        guide_path_prop.setVisible(False)
        psi_match_prop.setVisible(True)
        psi_arm_prop.setVisible(True)
        psi_mold_prop.setVisible(False)
        guide_view.camera().Elevation(-90)
        guide_view.camera().OrthogonalizeViewUp()
        guide_view.setCamera_FitAll()
        guide_view.snapshot_PNG(png)
        add_row_picture(doc_table, '快照（5/6）\nSnapshot', [s_body_center, s_body_center], png)

        aceta_prop.setVisible(False)
        guide_path_prop.setVisible(False)
        psi_match_prop.setVisible(True)
        psi_arm_prop.setVisible(True)
        psi_mold_prop.setVisible(True)
        update_cup_camera_plane()
        guide_view.camera().Elevation(-30)
        guide_view.camera().OrthogonalizeViewUp()
        guide_view.snapshot_PNG(png)
        add_row_picture(doc_table, '快照（6/6）\nSnapshot', [s_body_center, s_body_center], png)

        aceta_prop.setVisible(visibles[0])
        guide_path_prop.setVisible(visibles[1])
        psi_match_prop.setVisible(visibles[2])
        psi_arm_prop.setVisible(visibles[3])
        psi_mold_prop.setVisible(visibles[4])
        doc.save(file)


def LeftButtonPress(**kwargs):
    if kwargs['picked'] is voi_view:
        if plcut_box.text() == '请勾画切割范围':
            pt = voi_view.pickPt_FocalPlane()
            plcut_pts.clear()
            plcut_pts.append(pt)
            update_line_cut()


def LeftButtonPressMove(**kwargs):
    global spcut_center, pelvis_cs
    if kwargs['picked'] is voi_view:
        if plcut_box.text() == '请勾画切割范围':
            pt = voi_view.pickPt_Cell()

            for path_pt in plcut_pts:
                if (path_pt == pt).all():
                    return

            plcut_pts.append(pt)  # 添加拾取路径点
            update_line_cut()
        else:
            voi_view.mouseRotateFocal(**kwargs)
    elif kwargs['picked'] is pelvis_view:
        pelvis_view.mouseRotateFocal(**kwargs)
        if pelvis_MPP_box.text() == '确定':
            update_pelvis_MPP()
        elif pelvis_APP_box.text() == '确定':
            update_pelvis_APP()
    elif kwargs['picked'] in [spcut_prop, *spcut_y_props, *spcut_z_props]:
        spcut_center += kwargs['picked'].view().pickVt_FocalPlane()
        update_spcut()
        update_spcut_section()
    elif kwargs['picked'] in [cup_prop, cup_section_prop]:
        pt = pelvis_view.pickPt_FocalPlane()
        pt = vmi.ptOnPlane(pt, pelvis_cs.origin(), pelvis_view.cameraVt_Look())

        over = kwargs['picked'].view().pickVt_FocalPlane()
        if np.dot(pt - pelvis_cs.origin(), cup_cs.origin() - pelvis_cs.origin()) < 0:
            over = pelvis_cs.inv().mvt(over)
            over[0] *= -1
            over = pelvis_cs.mvt(over)

        cup_cs.setOrigin(cup_cs.origin() + over)
        update_cup()
        return True


def LeftButtonPressMoveRelease(**kwargs):
    if kwargs['picked'] is voi_view:
        if plcut_box.text() == '请勾画切割范围':
            plcut_box.draw_text('请耐心等待')
            plcut_voi_image()
            plcut_box.draw_text('线切割')

            plcut_pts.clear()
            update_line_cut()


def LeftButtonPressRelease(**kwargs):
    global spcut_radius, pelvis_cs, cup_radius, cup_cs, cup_RA, cup_RI, bone_value
    global body_angle, body_xr, body_yr, match_style
    if kwargs['picked'] is dicom_box:
        original_image = read_dicom()
        clear_guide_plate()
        if original_image is not None:
            original_slice.setData(original_image)
            original_slice.setSlicePlaneOrigin_Center()
            original_slice.setSlicePlaneNormal_Coronal()
            original_view.setCamera_Coronal()
            original_view.setCamera_FitAll()
    elif kwargs['picked'] is original_slice:
        on_init_voi()
    elif kwargs['picked'] is bone_value_box:
        v = vmi.askInt(-1000, bone_value, 3000)
        if v is not None:
            bone_value = v
            bone_value_box.draw_text('骨阈值：{:.0f} HU'.format(bone_value))
            voi_volume.setOpacityScalar({bone_value - 1: 0, bone_value: 1})
            clear_guide_plate()
    elif kwargs['picked'] is plcut_box:
        if plcut_box.text() == '线切割':
            plcut_box.draw_text('请勾画切割范围')
        elif plcut_box.text() == '请勾画切割范围':
            plcut_box.draw_text('线切割')
    elif kwargs['picked'] is spcut_box:
        spcut_box.draw_text('请耐心等待')
        spcut_voi_image()
        spcut_box.draw_text('球切割')
        update_spcut_section()
    elif kwargs['picked'] is spcut_visible_box:
        spcut_prop.visibleToggle()
        spcut_visible_box.draw_text(
            back_color=QColor().fromRgbF(0.4, 0.6, 1) if spcut_prop.visible() else QColor('black'))
    elif kwargs['picked'] is spcut_radius_box:
        r = vmi.askInt(5, spcut_radius, 50)
        if r is not None:
            spcut_radius = r
            spcut_radius_box.draw_text('球半径：{:.0f} mm'.format(spcut_radius))
            update_spcut()
            update_spcut_section()
    elif kwargs['picked'] is init_pelvis_box:
        init_pelvis_box.draw_text('请耐心等待')
        init_pelvis()
        update_cup_orientation()
        update_cup()
        init_pelvis_box.draw_text('创建骨盆')
    elif kwargs['picked'] is cup_radius_box:
        r = vmi.askInt(5, cup_radius, 50)
        if r is not None:
            cup_radius = r
            cup_radius_box.draw_text('半径：{:.0f} mm'.format(cup_radius))
            update_cup()
            clear_guide_plate()
    elif kwargs['picked'] is cup_RA_box:
        r = vmi.askFloat(-15.0, cup_RA, 45.0, 1, 0.1)
        if r is not None:
            cup_RA = r
            cup_RA_box.draw_text('前倾角：{:.1f}°'.format(cup_RA))

            update_cup_orientation()
            update_cup()
            clear_guide_plate()
    elif kwargs['picked'] is cup_RI_box:
        r = vmi.askFloat(0.0, cup_RI, 60.0, 1, 0.1)
        if r is not None:
            cup_RI = r
            cup_RI_box.draw_text('外展角：{:.1f}°'.format(cup_RI))

            update_cup_orientation()
            update_cup()
            clear_guide_plate()
    elif kwargs['picked'] is cup_camera_sagittal_box:
        update_cup_camera_sagittal()
    elif kwargs['picked'] is cup_camera_coronal_box:
        update_cup_camera_coronal()
    elif kwargs['picked'] is cup_camera_axial_box:
        update_cup_camera_axial()
    elif kwargs['picked'] is cup_camera_RA_box:
        update_cup_camera_RA()
    elif kwargs['picked'] is cup_camera_RI_box:
        update_cup_camera_RI()
    elif kwargs['picked'] is cup_camera_plane_box:
        update_cup_camera_plane()
    elif kwargs['picked'] is pelvis_visible_box:
        pelvis_prop.visibleToggle()
        pelvis_visible_box.draw_text(
            back_color=QColor().fromRgbF(1, 1, 0.6) if pelvis_prop.visible() else QColor('black'))
    elif kwargs['picked'] is metal_visible_box:
        metal_prop.visibleToggle()
        metal_visible_box.draw_text(
            back_color=QColor().fromRgbF(1, 0.4, 0.4) if metal_prop.visible() else QColor('black'))
    elif kwargs['picked'] is pelvis_slice_box:
        if pelvis_slice_box.text() == '三维':
            pelvis_slice_box.draw_text('断层')
            cup_prop.setVisible(False)
            pelvis_prop.setVisible(False)
            cup_section_prop.setVisible(True)
            pelvis_slice.setVisible(True)

            pelvis_slice.setSlicePlane(cup_cs.origin(), pelvis_view.cameraVt_Look())
            pd = vmi.pdCut_Implicit(cup_prop.data(), pelvis_slice.slicePlane())
            cup_section_prop.setData(pd)
        elif pelvis_slice_box.text() == '断层':
            pelvis_slice_box.draw_text('三维')
            cup_prop.setVisible(True)
            pelvis_prop.setVisible(True)
            cup_section_prop.setVisible(False)
            pelvis_slice.setVisible(False)
    elif kwargs['picked'] is pelvis_MPP_box:
        if pelvis_MPP_box.text() == '配准对称面':
            pelvis_MPP_box.draw_text('确定')
            update_pelvis_MPP()
            pelvis_MPP_prop.setVisible(True)
            pelvis_cs_indicator_prop.setVisible(True)
        elif pelvis_MPP_box.text() == '确定':
            pelvis_MPP_box.draw_text('配准对称面')
            pelvis_MPP_prop.setVisible(False)
            pelvis_cs_indicator_prop.setVisible(False)
    elif kwargs['picked'] is pelvis_APP_box:
        if pelvis_APP_box.text() == '配准前平面':
            pelvis_APP_box.draw_text('确定')
            update_pelvis_APP()
            pelvis_cs_indicator_prop.setVisible(True)
        elif pelvis_APP_box.text() == '确定':
            pelvis_APP_box.draw_text('配准前平面')
            pelvis_cs_indicator_prop.setVisible(False)
    elif kwargs['picked'] is init_guide_box:
        init_guide_box.draw_text('请耐心等待')
        init_guide()
        init_guide_box.draw_text('创建导板')
    elif kwargs['picked'] is case_sync_box:
        case_sync_box.draw_text('请耐心等待')
        case_sync()
        case_sync_box.draw_text('同步')
    elif kwargs['picked'] is aceta_visible_box:
        aceta_prop.visibleToggle()
        aceta_visible_box.draw_text(
            back_color=QColor().fromRgbF(1, 1, 0.6) if aceta_prop.visible() else QColor('black'))
    elif kwargs['picked'] is psi_match_visible_box:
        psi_match_prop.visibleToggle()
        psi_match_visible_box.draw_text(
            back_color=QColor().fromRgbF(0.4, 0.6, 1) if psi_match_prop.visible() else QColor('black'))
    elif kwargs['picked'] is guide_path_visible_box:
        guide_path_prop.visibleToggle()
        guide_path_visible_box.draw_text(
            back_color=QColor().fromRgbF(0.4, 1, 0.4) if guide_path_prop.visible() else QColor('black'))
    elif kwargs['picked'] is psi_arm_visible_box:
        psi_arm_prop.visibleToggle()
        psi_arm_visible_box.draw_text(
            back_color=QColor().fromRgbF(0.6, 0.8, 1) if psi_arm_prop.visible() else QColor('black'))
    elif kwargs['picked'] is mold_visible_box:
        psi_mold_prop.visibleToggle()
        mold_visible_box.draw_text(
            back_color=QColor().fromRgbF(1, 0.4, 0.4) if psi_mold_prop.visible() else QColor('black'))
    elif kwargs['picked'] is guide_update_box:
        guide_update_box.draw_text('请耐心等待')
        update_guide_path()
        update_guide_plate()
        guide_update_box.draw_text('更新导板')
    elif kwargs['picked'] is body_angle_box:
        r = vmi.askInt(-45, body_angle, 45)
        if r is not None:
            body_angle = r
            body_angle_box.draw_text('主体方位：{}°'.format(body_angle))
            update_guide_path()
            clear_guide_plate()
    elif kwargs['picked'] is body_xr_box:
        r = vmi.askInt(10, body_xr, 50)
        if r is not None:
            body_xr = r
            body_xr_box.draw_text('主体长径：{:.0f} mm'.format(body_xr))
            update_guide_path()
            clear_guide_plate()
    elif kwargs['picked'] is body_yr_box:
        r = vmi.askInt(10, body_yr, 50)
        if r is not None:
            body_yr = r
            body_yr_box.draw_text('主体宽径：{:.0f} mm'.format(body_yr))
            update_guide_path()
            clear_guide_plate()
    elif kwargs['picked'] is arm_length_box:
        global arm_length
        r = vmi.askInt(10, arm_length, 90)
        if r is not None:
            arm_length = r
            arm_length_box.draw_text('旋臂长度：{:.0f} mm'.format(arm_length))
            update_guide_path()
            clear_guide_plate()
    elif kwargs['picked'] is match_style_box:
        match_style = '上浮式' if match_style == '下沉式' else '下沉式'
        match_style_box.draw_text('匹配方式：{}'.format(match_style))
        update_guide_path()
        clear_guide_plate()


def MidButtonPressMove(**kwargs):
    if kwargs['picked'] in [pelvis_view, pelvis_slice]:
        pelvis_view.mousePan()
        if pelvis_MPP_box.text() == '确定':
            update_pelvis_MPP()
        elif pelvis_APP_box.text() == '确定':
            update_pelvis_APP()


def RightButtonPressMove(**kwargs):
    if kwargs['picked'] in [pelvis_view, pelvis_slice]:
        pelvis_view.mouseZoom()
        if pelvis_MPP_box.text() == '确定':
            update_pelvis_MPP()
        elif pelvis_APP_box.text() == '确定':
            update_pelvis_APP()


def NoButtonWheel(**kwargs):
    global bone_value, spcut_radius, cup_radius, cup_RA, cup_RI
    global body_angle, body_xr, body_yr, kirs_radius
    if kwargs['picked'] is bone_value_box:
        bone_value = min(max(bone_value + 10 * kwargs['delta'], -1000), 3000)
        bone_value_box.draw_text('骨阈值：{:.0f} HU'.format(bone_value))
        voi_volume.setOpacityScalar({bone_value - 1: 0, bone_value: 1})
        clear_guide_plate()
    elif kwargs['picked'] is spcut_radius_box:
        spcut_radius = min(max(spcut_radius + kwargs['delta'], 5), 50)
        spcut_radius_box.draw_text('球半径：{:.0f} mm'.format(spcut_radius))
        update_spcut()
        update_spcut_section()
    elif kwargs['picked'] is pelvis_view:
        pelvis_view.mouseRotateLook(**kwargs)
        if pelvis_MPP_box.text() == '确定':
            update_pelvis_MPP()
        elif pelvis_APP_box.text() == '确定':
            update_pelvis_APP()
    elif kwargs['picked'] is cup_radius_box:
        cup_radius = min(max(cup_radius + kwargs['delta'], 5), 50)
        cup_radius_box.draw_text('半径：{:.0f} mm'.format(cup_radius))
        update_cup()
        clear_guide_plate()
    elif kwargs['picked'] is cup_RA_box:
        cup_RA = min(max(cup_RA + 0.1 * kwargs['delta'], -15), 45)
        cup_RA_box.draw_text('前倾角：{:.1f}°'.format(cup_RA))
        update_cup_orientation()
        update_cup()
        clear_guide_plate()
    elif kwargs['picked'] is cup_RI_box:
        cup_RI = min(max(cup_RI + 0.1 * kwargs['delta'], 0), 60)
        cup_RI_box.draw_text('外展角：{:.1f}°'.format(cup_RI))
        update_cup_orientation()
        update_cup()
        clear_guide_plate()
    elif kwargs['picked'] is pelvis_slice:
        update_cup()
    elif kwargs['picked'] is body_angle_box:
        body_angle = min(max(body_angle + kwargs['delta'], -45), 45)
        body_angle_box.draw_text('主体方位：{}°'.format(body_angle))
        update_guide_path()
        clear_guide_plate()
    elif kwargs['picked'] is body_xr_box:
        body_xr = min(max(body_xr + kwargs['delta'], 10), 50)
        body_xr_box.draw_text('主体长径：{:.0f} mm'.format(body_xr))
        update_guide_path()
        clear_guide_plate()
    elif kwargs['picked'] is body_yr_box:
        body_yr = min(max(body_yr + kwargs['delta'], 10), 50)
        body_yr_box.draw_text('主体宽径：{:.0f} mm'.format(body_yr))
        update_guide_path()
        clear_guide_plate()
    elif kwargs['picked'] is arm_length_box:
        global arm_length
        arm_length = min(max(arm_length + kwargs['delta'], 10), 90)
        arm_length_box.draw_text('旋臂长度：{:.0f} mm'.format(arm_length))
        update_guide_path()
        clear_guide_plate()


def return_globals():
    return globals()


if __name__ == '__main__':
    global original_view, voi_view, spcut_y_views, spcut_z_views, pelvis_view, guide_view
    main = vmi.Main(return_globals)
    main.setAppName('Propsi TH')
    main.setAppVersion('f5dbde2e7f78c3134da1bcd4f7118916fd810d91')
    main.excludeKeys += ['main', 'i', 'box', 'prop', 'voi_image']

    select = vmi.askButtons(['新建', '打开'], title=main.appName())
    if select is None:
        vmi.appexit()

    if select == '打开':
        open_file = vmi.askOpenFile(nameFilter='*.vmi', title='打开存档')
        if open_file is not None:
            main.loads(pathlib.Path(open_file).read_bytes())
        else:
            vmi.appexit()
    elif select == '新建':
        patient_name = str()

        # 0 原始图像
        original_target, LR = np.zeros(3), 1
        original_view = vmi.View()

        original_slice = vmi.ImageSlice(original_view)  # 断层显示
        original_slice.setColorWindow_Bone()
        original_slice.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        dicom_box = vmi.TextBox(original_view, text='DICOM', pickable=True,
                                size=[0.2, 0.04], pos=[0, 0.04], anchor=[0, 0])
        dicom_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        patient_name_box = vmi.TextBox(original_view, text='姓名', visible=False,
                                       size=[0.2, 0.04], pos=[0, 0.1], anchor=[0, 0])
        patient_LR_box = vmi.TextBox(original_view, text='患侧', visible=False,
                                     size=[0.2, 0.04], pos=[0, 0.14], anchor=[0, 0])

        # 1 目标区域
        voi_view = vmi.View()
        voi_view.mouse['LeftButton']['Press'] = [LeftButtonPress]
        voi_view.mouse['LeftButton']['PressMove'] = [LeftButtonPressMove]
        voi_view.mouse['LeftButton']['PressMoveRelease'] = [LeftButtonPressMoveRelease]

        bone_value, target_value = 200, -1100

        bone_value_box = vmi.TextBox(voi_view, text='骨阈值：{:.0f} HU'.format(bone_value),
                                     size=[0.2, 0.04], pos=[0, 0.04], anchor=[0, 0], pickable=True)
        bone_value_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]
        bone_value_box.mouse['NoButton']['Wheel'] = [NoButtonWheel]

        voi_size, voi_center, voi_origin, voi_cs = np.zeros(3), np.zeros(3), np.zeros(3), vmi.CS4x4()
        voi_image = vtk.vtkImageData()

        voi_volume = vmi.ImageVolume(voi_view, pickable=True)
        voi_volume.setOpacityScalar({bone_value - 1: 0, bone_value: 1})
        voi_volume.setColor({bone_value: [1, 1, 0.6]})

        # 线切割
        plcut_pts = []
        plcut_prop = vmi.PolyActor(voi_view, color=[1, 0.6, 0.6], line_width=3, always_on_top=True)

        # 球切割
        spcut_center, spcut_radius, spcut_length = np.zeros(3), 22, 200
        spcut_prop = vmi.PolyActor(voi_view, color=[0.4, 0.6, 1], pickable=True)
        spcut_prop.mouse['LeftButton']['PressMove'] = [LeftButtonPressMove]

        plcut_box = vmi.TextBox(voi_view, text='线切割', size=[0.2, 0.04], pos=[0, 0.1], anchor=[0, 0], pickable=True)
        plcut_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        spcut_box = vmi.TextBox(voi_view, text='球切割', size=[0.2, 0.04], pos=[0, 0.16], anchor=[0, 0], pickable=True)
        spcut_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        spcut_visible_box = vmi.TextBox(voi_view, size=[0.01, 0.04], pos=[0.2, 0.16], anchor=[1, 0],
                                        back_color=QColor().fromRgbF(0.4, 0.6, 1), pickable=True)
        spcut_visible_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        spcut_radius_box = vmi.TextBox(voi_view, text='球半径：{:.0f} mm'.format(spcut_radius),
                                       size=[0.2, 0.04], pos=[0, 0.20], anchor=[0, 0], pickable=True)
        spcut_radius_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]
        spcut_radius_box.mouse['NoButton']['Wheel'] = [NoButtonWheel]

        # 234 球切割局部
        spcut_y_views = [vmi.View(), vmi.View(), vmi.View()]
        spcut_z_views = [vmi.View(), vmi.View(), vmi.View()]

        for i in spcut_y_views + spcut_z_views:
            i.mouse['LeftButton']['PressMove'] = [i.mouseBlock]

        spcut_y_boxs = [vmi.TextBox(spcut_y_views[i], text=['斜冠状位', '正冠状位', '斜冠状位'][i],
                                    size=[0.2, 0.1], pos=[1, 1], anchor=[1, 1]) for i in range(3)]
        spcut_z_boxs = [vmi.TextBox(spcut_z_views[i], text=['斜横断位', '正横断位', '斜横断位'][i],
                                    size=[0.2, 0.1], pos=[1, 1], anchor=[1, 1]) for i in range(3)]

        spcut_y_slices = [vmi.ImageSlice(v) for v in spcut_y_views]
        spcut_z_slices = [vmi.ImageSlice(v) for v in spcut_z_views]

        for i in spcut_y_slices + spcut_z_slices:
            i.bindData(voi_volume)
            i.mouse['NoButton']['Wheel'] = [i.mouseBlock]

        spcut_y_props = [vmi.PolyActor(v, color=[0.4, 0.6, 1], opacity=0.5, pickable=True) for v in spcut_y_views]
        spcut_z_props = [vmi.PolyActor(v, color=[0.4, 0.6, 1], opacity=0.5, pickable=True) for v in spcut_z_views]

        for i in spcut_y_props + spcut_z_props:
            i.mouse['LeftButton']['PressMove'] = [LeftButtonPressMove]

        init_pelvis_box = vmi.TextBox(voi_view, text='创建骨盆', fore_color=QColor('white'), back_color=QColor('crimson'),
                                      bold=True, size=[0.2, 0.04], pos=[1, 0.04], anchor=[1, 0], pickable=True)
        init_pelvis_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        # 5 骨盆规划
        pelvis_view = vmi.View()
        pelvis_view.mouse['LeftButton']['PressMove'] = [LeftButtonPressMove]
        pelvis_view.mouse['MidButton']['PressMove'] = [MidButtonPressMove]
        pelvis_view.mouse['RightButton']['PressMove'] = [RightButtonPressMove]
        pelvis_view.mouse['NoButton']['Wheel'] = [NoButtonWheel]

        pelvis_prop = vmi.PolyActor(pelvis_view, color=[1, 1, 0.6])
        metal_prop = vmi.PolyActor(pelvis_view, color=[1, 0.4, 0.4])

        pelvis_cs, cup_cs = vmi.CS4x4(), vmi.CS4x4()
        cup_radius, cup_RA, cup_RI = 22, 20, 40
        cup_prop = vmi.PolyActor(pelvis_view, color=[0.4, 0.6, 1], line_width=3, pickable=True)
        cup_prop.mouse['LeftButton']['PressMove'] = [LeftButtonPressMove]

        cup_radius_box = vmi.TextBox(pelvis_view, text='半径：{:.0f} mm'.format(cup_radius), pickable=True,
                                     size=[0.2, 0.04], pos=[0, 0.04], anchor=[0, 0])
        cup_radius_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]
        cup_radius_box.mouse['NoButton']['Wheel'] = [NoButtonWheel]

        cup_RA_box = vmi.TextBox(pelvis_view, text='前倾角：{:.1f}°'.format(cup_RA), pickable=True,
                                 size=[0.2, 0.04], pos=[0, 0.1], anchor=[0, 0])
        cup_RA_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]
        cup_RA_box.mouse['NoButton']['Wheel'] = [NoButtonWheel]

        cup_RI_box = vmi.TextBox(pelvis_view, text='外展角：{:.1f}°'.format(cup_RI), pickable=True,
                                 size=[0.2, 0.04], pos=[0, 0.14], anchor=[0, 0])
        cup_RI_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]
        cup_RI_box.mouse['NoButton']['Wheel'] = [NoButtonWheel]

        pelvis_slice = vmi.ImageSlice(pelvis_view, visible=False)
        pelvis_slice.bindData(voi_volume)
        pelvis_slice.mouse['MidButton']['PressMove'] = [MidButtonPressMove]
        pelvis_slice.mouse['RightButton']['PressMove'] = [RightButtonPressMove]
        pelvis_slice.mouse['NoButton']['Wheel'] += [NoButtonWheel]

        original_slice.bindColorWindow(spcut_y_slices + spcut_z_slices + [pelvis_slice])

        cup_section_prop = vmi.PolyActor(pelvis_view, color=[0.4, 0.6, 1], line_width=3, visible=False, pickable=True)
        cup_section_prop.mouse['LeftButton']['PressMove'] = [LeftButtonPressMove]

        pelvis_slice_box = vmi.TextBox(pelvis_view, text='三维', pickable=True,
                                       size=[0.2, 0.04], pos=[0, 0.2], anchor=[0, 0])
        pelvis_slice_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        pelvis_MPP_box = vmi.TextBox(pelvis_view, text='配准对称面', pickable=True,
                                     size=[0.2, 0.04], pos=[0, 0.26], anchor=[0, 0])
        pelvis_MPP_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        pelvis_APP_box = vmi.TextBox(pelvis_view, text='配准前平面', pickable=True,
                                     size=[0.2, 0.04], pos=[0, 0.3], anchor=[0, 0])
        pelvis_APP_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        cup_camera_sagittal_box = vmi.TextBox(pelvis_view, text='矢状', pickable=True,
                                              size=[0.06, 0.04], pos=[0, 0.36], anchor=[0, 0])
        cup_camera_sagittal_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        cup_camera_coronal_box = vmi.TextBox(pelvis_view, text='冠状', pickable=True,
                                             size=[0.06, 0.04], pos=[0.06, 0.36], anchor=[0, 0])
        cup_camera_coronal_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        cup_camera_axial_box = vmi.TextBox(pelvis_view, text='横断', pickable=True,
                                           size=[0.06, 0.04], pos=[0.12, 0.36], anchor=[0, 0])
        cup_camera_axial_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        cup_camera_RA_box = vmi.TextBox(pelvis_view, text='前倾', pickable=True,
                                        size=[0.06, 0.04], pos=[0, 0.4], anchor=[0, 0])
        cup_camera_RA_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        cup_camera_RI_box = vmi.TextBox(pelvis_view, text='外展', pickable=True,
                                        size=[0.06, 0.04], pos=[0.06, 0.4], anchor=[0, 0])
        cup_camera_RI_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        cup_camera_plane_box = vmi.TextBox(pelvis_view, text='正视', pickable=True,
                                           size=[0.06, 0.04], pos=[0.12, 0.4], anchor=[0, 0])
        cup_camera_plane_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        pelvis_visible_box = vmi.TextBox(pelvis_view, size=[0.02, 0.04], pos=[0.18, 0.36], anchor=[0, 0],
                                         back_color=QColor().fromRgbF(1, 1, 0.6), pickable=True)
        pelvis_visible_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        metal_visible_box = vmi.TextBox(pelvis_view, size=[0.02, 0.04], pos=[0.18, 0.4], anchor=[0, 0],
                                        back_color=QColor().fromRgbF(1, 0.4, 0.4), pickable=True)
        metal_visible_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        pelvis_MPP_prop = vmi.PolyActor(pelvis_view, visible=False, color=[1, 0.4, 0.4], repres='points')
        pelvis_cs_indicator_prop = vmi.PolyActor(pelvis_view, visible=False, color=[1, 0.4, 0.4], line_width=3,
                                                 always_on_top=True)

        init_guide_box = vmi.TextBox(pelvis_view, text='创建导板', pickable=True,
                                     fore_color=QColor('white'), back_color=QColor('crimson'),
                                     bold=True, size=[0.2, 0.04], pos=[1, 0.04], anchor=[1, 0])
        init_guide_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        # 8 导板
        guide_view = vmi.View()
        guide_view.bindCamera([pelvis_view])

        aceta_prop = vmi.PolyActor(guide_view, color=[1, 1, 0.6])

        guide_path = vmi.TopoDS_Shape()  # 引导线
        guide_path_prop = vmi.PolyActor(guide_view, color=[0.4, 1, 0.4], line_width=3)

        psi_match = vmi.TopoDS_Shape()  # 主板
        psi_match_prop = vmi.PolyActor(guide_view, color=[0.4, 0.6, 1])

        psi_arm = vmi.TopoDS_Shape()  # 旋臂
        psi_arm_prop = vmi.PolyActor(guide_view, color=[0.6, 0.8, 1])
        psi_arm_prop.mouse['LeftButton']['PressMove'] = [LeftButtonPressMove]

        psi_mold = vmi.TopoDS_Shape()  # 配模
        psi_mold_prop = vmi.PolyActor(guide_view, color=[1, 0.4, 0.4])

        body_angle = 0  # 主体方位
        body_xr, body_yr = 20, 15  # 主体长径宽径，臼窝匹配块
        arm_length = 50  # 旋臂长度
        match_style = '下沉式'
        kirs_radius = 1  # 导向孔半径

        body_angle_box = vmi.TextBox(guide_view, text='主体方位：{}°'.format(body_angle), pickable=True,
                                     size=[0.2, 0.04], pos=[0, 0.04], anchor=[0, 0])
        body_angle_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]
        body_angle_box.mouse['NoButton']['Wheel'] = [NoButtonWheel]

        body_xr_box = vmi.TextBox(guide_view, text='主体长径：{:.0f} mm'.format(body_xr), pickable=True,
                                  size=[0.2, 0.04], pos=[0, 0.08], anchor=[0, 0])
        body_xr_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]
        body_xr_box.mouse['NoButton']['Wheel'] = [NoButtonWheel]

        body_yr_box = vmi.TextBox(guide_view, text='主体宽径：{:.0f} mm'.format(body_yr), pickable=True,
                                  size=[0.2, 0.04], pos=[0, 0.12], anchor=[0, 0])
        body_yr_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]
        body_yr_box.mouse['NoButton']['Wheel'] = [NoButtonWheel]

        arm_length_box = vmi.TextBox(guide_view, text='旋臂长度：{:.0f} mm'.format(arm_length), pickable=True,
                                     size=[0.2, 0.04], pos=[0, 0.16], anchor=[0, 0])
        arm_length_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]
        arm_length_box.mouse['NoButton']['Wheel'] = [NoButtonWheel]

        match_style_box = vmi.TextBox(guide_view, text='匹配方式：{}'.format(match_style), pickable=True,
                                      size=[0.2, 0.04], pos=[0, 0.20], anchor=[0, 0])
        match_style_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        guide_update_box = vmi.TextBox(guide_view, text='更新导板', pickable=True,
                                       size=[0.10, 0.04], pos=[0, 0.26], anchor=[0, 0])
        guide_update_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        aceta_visible_box = vmi.TextBox(guide_view, pickable=True, back_color=QColor().fromRgbF(1, 1, 0.6),
                                        size=[0.02, 0.04], pos=[0.10, 0.26], anchor=[0, 0])
        aceta_visible_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        guide_path_visible_box = vmi.TextBox(guide_view, pickable=True, back_color=QColor().fromRgbF(0.4, 1, 0.4),
                                             size=[0.02, 0.04], pos=[0.12, 0.26], anchor=[0, 0])
        guide_path_visible_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        psi_match_visible_box = vmi.TextBox(guide_view, pickable=True, back_color=QColor().fromRgbF(0.4, 0.6, 1),
                                             size=[0.02, 0.04], pos=[0.14, 0.26], anchor=[0, 0])
        psi_match_visible_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        psi_arm_visible_box = vmi.TextBox(guide_view, pickable=True, back_color=QColor().fromRgbF(0.6, 0.8, 1),
                                            size=[0.02, 0.04], pos=[0.16, 0.26], anchor=[0, 0])
        psi_arm_visible_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        mold_visible_box = vmi.TextBox(guide_view, pickable=True, back_color=QColor().fromRgbF(1, 0.4, 0.4),
                                       size=[0.02, 0.04], pos=[0.18, 0.26], anchor=[0, 0])
        mold_visible_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        case_sync_box = vmi.TextBox(guide_view, text='同步', pickable=True,
                                    fore_color=QColor('white'), back_color=QColor('crimson'),
                                    bold=True, size=[0.2, 0.04], pos=[1, 0.04], anchor=[1, 0])
        case_sync_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        # output_box = vmi.TextBox(guide_view, text='导出方案', fore_color=QColor('white'), back_color=QColor('crimson'),
        #                          bold=True, size=[0.2, 0.04], pos=[1, 0.04], anchor=[1, 0], pickable=True)
        # output_box.mouse['LeftButton']['PressRelease'] = [LeftButtonPressRelease]

        for v in [original_view, voi_view, *spcut_y_views, *spcut_z_views, pelvis_view, guide_view]:
            v.setEnabled(False)
        original_view.setEnabled(True)
        guide_view.setEnabled(True)

    # 视图布局
    for v in [original_view, voi_view, *spcut_y_views, *spcut_z_views, pelvis_view, guide_view]:
        v.setMinimumWidth(round(0.5 * main.screenWidth()))

    for v in spcut_y_views + spcut_z_views:
        v.setMinimumWidth(round(0.25 * main.screenWidth()))

    original_view.setParent(main.scrollArea())
    main.layout().addWidget(original_view, 0, 0, 3, 1)
    main.layout().addWidget(voi_view, 0, 1, 3, 1)
    main.layout().addWidget(spcut_y_views[0], 0, 2, 1, 1)
    main.layout().addWidget(spcut_y_views[1], 1, 2, 1, 1)
    main.layout().addWidget(spcut_y_views[2], 2, 2, 1, 1)
    main.layout().addWidget(spcut_z_views[0], 0, 3, 1, 1)
    main.layout().addWidget(spcut_z_views[1], 1, 3, 1, 1)
    main.layout().addWidget(spcut_z_views[2], 2, 3, 1, 1)
    main.layout().addWidget(pelvis_view, 0, 4, 3, 1)
    main.layout().addWidget(guide_view, 0, 5, 3, 1)

    vmi.appexec(main)  # 执行主窗口程序
    vmi.appexit()  # 清理并退出程序
